# Copyright (C) 2013-2016 DNAnexus, Inc.
# -*- coding: utf-8 -*-
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""
Workflow Builder Library
+++++++++++++++++++

Contains utility methods useful for deploying workflows onto the platform.
"""

from __future__ import print_function, unicode_literals, division, absolute_import
import os
import sys
import json
import copy

import dxpy
from .cli import INTERACTIVE_CLI
from .cli.parsers import process_extra_args
from .utils.printing import fill
from .utils import json_load_raise_on_duplicates
from .exceptions import err_exit
from . import logger

UPDATABLE_GLOBALWF_FIELDS = {'title', 'summary', 'description', 'developerNotes', 'details'}
GLOBALWF_SUPPORTED_KEYS = UPDATABLE_GLOBALWF_FIELDS.union({"name", "version", "regionalOptions",
                                                           "categories", "billTo", "dxapi", "tags"})
SUPPORTED_KEYS = GLOBALWF_SUPPORTED_KEYS.union({"project", "folder", "outputFolder", "stages",
                                                "inputs", "outputs", "ignoreReuse", "properties"})

class WorkflowBuilderException(Exception):
    """
    This exception is raised by the methods in this module when workflow
    building fails.
    """
    pass


def _fetch_spec_from_dxworkflowjson(src_dir, json_file_name, parser):
    """
    Returns the parsed contents of a json specification.
    Raises WorkflowBuilderException (exit code 3) if this cannot be done.
    """
    if not os.path.isdir(src_dir):
        parser.error("{} is not a directory".format(src_dir))

    if not os.path.exists(os.path.join(src_dir, json_file_name)):
        raise WorkflowBuilderException(
            "Directory {} does not contain dxworkflow.json: not a valid DNAnexus workflow source directory"
            .format(src_dir))

    with open(os.path.join(src_dir, json_file_name)) as desc:
        try:
            return json_load_raise_on_duplicates(desc)
        except Exception as e:
            raise WorkflowBuilderException("Could not parse {} file as JSON: {}".format(json_file_name, e.args))

def _cleanup_empty_keys(json_spec):
    import re
    clean_json = re.sub('\"\w*\": (null|\{\}|\"\"|\[\])(\,|)\s*','',json.dumps(json_spec)).replace(", }","}")
    return json.loads(clean_json)

def _check_dxcompiler_version(json_spec):
    SUPPORTED_DXCOMPILER_VERSION = "2.8.0"
    try:
        from distutils.version import StrictVersion as Version
    except ImportError:
        from .utils.version import Version
    if json_spec.get("details") and json_spec["details"].get("version"):
        compiler_version_used = str(json_spec["details"].get("version"))
        compiler_version_no_snapshot = compiler_version_used.split("-")[0]
        if Version(compiler_version_no_snapshot) < Version(SUPPORTED_DXCOMPILER_VERSION):
            raise WorkflowBuilderException("Source workflow {} is not compiled using dxCompiler (version>={}) that supports creating global workflows.".format(json_spec["name"], SUPPORTED_DXCOMPILER_VERSION))
    else:
        raise WorkflowBuilderException("Cannot find the dxCompiler version from the dxworkflow.json/source workflow spec. Please specify it by updating the details field of the dxworkflow.json/source workflow spec using the 'version' key.")

def _notify_instance_type_selection(json_spec):
    is_static = json_spec["details"].get("staticInstanceTypeSelection", False)
    if is_static:
        print("Note: {workflow} was compiled with -instanceTypeSelection=static "
              "and will produce a global workflow that relies on instance types that may not be available to all users."
              .format(workflow=json_spec["name"]))

def _notify_dependencies(json_spec):
    # APPS-622: A textual representation of all the dependencies is appended to 
    # the "description" field of every workflow generated by the dxCompiler in markdown format.
     dependency_list = json_spec.get("description")
     if dependency_list:
         print("Note: {workflow} was compiled with unbundled dependencies. "
              "Please check the workflow description and make sure access to these dependencies is provided to all authorized users."
              .format(workflow=json_spec["name"]))

def _get_destination_project(json_spec, args, build_project_id=None):
    """
    Returns destination project in which the workflow should be created.
    In can be set in multiple ways whose order of precedence is:
    1. --destination, -d option supplied with `dx build`,
    2. 'project' specified in the json file,
    3. project set in the dxpy.WORKSPACE_ID environment variable.
    """
    if build_project_id:
        return build_project_id
    if 'project' in json_spec:
        return json_spec['project']
    if dxpy.WORKSPACE_ID:
        return dxpy.WORKSPACE_ID
    error_msg = "Can't create a workflow without specifying a destination project; "
    error_msg += "please use the -d/--destination flag to explicitly specify a project"
    raise WorkflowBuilderException(error_msg)


def _get_destination_folder(json_spec, folder_name=None):
    """
    Returns destination project in which the workflow should be created.
    It can be set in the json specification or by --destination option supplied
    with `dx build`.
    The order of precedence is:
    1. --destination, -d option,
    2. 'folder' specified in the json file.
    """
    dest_folder = folder_name or json_spec.get('folder') or '/'
    if not dest_folder.endswith('/'):
        dest_folder = dest_folder + '/'
    return dest_folder


def _get_workflow_name(json_spec, workflow_name=None):
    """
    Returns the name of the workflow to be created. It can be set in the json
    specification or by --destination option supplied with `dx build`.
    The order of precedence is:
    1. --destination, -d option,
    2. 'name' specified in the json file.
    If not provided, returns empty string.
    """
    return workflow_name or json_spec.get('name')


def _get_unsupported_keys(keys, supported_keys):
    return [key for key in keys if key not in supported_keys]


def _set_categories_on_workflow(global_workflow_id, categories_to_set):
    """
    Note: Categories are set on the workflow series level,
    i.e. the same set applies to all versions.
    """
    assert(isinstance(categories_to_set, list))

    existing_categories = dxpy.api.global_workflow_list_categories(global_workflow_id)['categories']
    categories_to_add = set(categories_to_set).difference(set(existing_categories))
    categories_to_remove = set(existing_categories).difference(set(categories_to_set))
    if categories_to_add:
        dxpy.api.global_workflow_add_categories(global_workflow_id,
                                                input_params={'categories': list(categories_to_add)})
    if categories_to_remove:
        dxpy.api.global_workflow_remove_categories(global_workflow_id,
                                                   input_params={'categories': list(categories_to_remove)})


def _version_exists(json_spec, name=None, version=None):
    """
    Returns True if a global workflow with the given name and version
    already exists in the platform and the user has developer rights
    to the workflow. "name" and "version" can be passed if we already
    made a "describe" API call on the global workflow and so know the
    requested name and version already exists.
    """
    try: 
        requested_name = json_spec['name']
        requested_version = json_spec['version']
    except:
        raise WorkflowBuilderException("Both 'name' and 'version' fields must be given in the dxworkflow.json/source workflow spec to build/update a global workflow.")

    if requested_name == name and requested_version == version:
        return True
    else:
        try:
            desc_output = dxpy.api.global_workflow_describe('globalworkflow-' + json_spec['name'],
                                                            alias=json_spec['version'],
                                                            input_params={"fields": {"name": True,
                                                                                     "version": True}})
            return desc_output['name'] == json_spec['name'] and desc_output['version'] == json_spec['version']
        except dxpy.exceptions.DXAPIError:
            return False
        except:
            raise


def _get_validated_stage(stage, stage_index):
    # required keys
    if 'executable' not in stage:
        raise WorkflowBuilderException(
            "executable is not specified for stage with index {}".format(stage_index))

    # print ignored keys if present in json_spec
    supported_keys = {"id", "input", "executable", "name", "folder",
                      "executionPolicy", "systemRequirements"}
    unsupported_keys = _get_unsupported_keys(stage.keys(), supported_keys)
    if len(unsupported_keys) > 0:
        print("Warning: the following stage fields are not supported and will be ignored: {}"
              .format(", ".join(unsupported_keys)))

    return stage


def validate_ignore_reuse(stages, ignore_reuse_stages):
    """
    Checks if each stage ID specified in ignore_reuse_stages exists in
    the workflow definition. If ignore_reuse_stages contains
    only '*', the field is valid.
    """
    if not isinstance(ignore_reuse_stages, list):
         raise WorkflowBuilderException('"IgnoreReuse must be a list of strings - stage IDs or "*"')

    ignore_reuse_set = set(ignore_reuse_stages)
    if '*' in ignore_reuse_set and len(ignore_reuse_set) == 1:
        return

    stage_ids = set([stage.get('id') for stage in stages])
    for ignored in ignore_reuse_set:
        if ignored not in stage_ids:
            raise WorkflowBuilderException(
                'Stage with ID {} not found. Add a matching "id" for the stage you wish to set ignoreReuse for'.format(ignored))


def _get_validated_stages(stages):
    """
    Validates stages of the workflow as a list of dictionaries.
    """
    if not isinstance(stages, list):
        raise WorkflowBuilderException("Stages must be specified as a list of dictionaries")
    validated_stages = []
    for index, stage in enumerate(stages):
        validated_stages.append(_get_validated_stage(stage, index))
    return validated_stages


def _validate_json_for_regular_workflow(json_spec, args):
    """
    Validates fields used only for building a regular, project-based workflow.
    """
    validated = {}
    override_project_id, override_folder, override_workflow_name = \
        dxpy.executable_builder.get_parsed_destination(args.destination)
    validated['project'] = _get_destination_project(json_spec, args, override_project_id)
    validated['folder'] = _get_destination_folder(json_spec, override_folder)

    workflow_name = _get_workflow_name(json_spec, override_workflow_name)
    if not workflow_name:
        print('Warning: workflow name is not specified')
    else:
        validated['name'] = workflow_name
    return validated


def _validate_json_for_global_workflow(json_spec, args):
    """
    Validates fields used for building a global workflow.
    Since building a global workflow is done after all the underlying workflows
    are built, which may be time-consuming, we validate as much as possible here.
    """
    if 'name' not in json_spec:
        raise WorkflowBuilderException(
            "dxworkflow.json/source workflow spec contains no 'name' field, but it is required to build a global workflow")
    if not dxpy.executable_builder.GLOBAL_EXEC_NAME_RE.match(json_spec['name']):
        raise WorkflowBuilderException(
            "The name of your workflow must match /^[a-zA-Z0-9._-]+$/")
    if json_spec['name'] != json_spec['name'].lower():
        logger.warn('workflow name "{}" should be all lowercase'.format(json_spec['name']))

    if 'version' not in json_spec:
        raise WorkflowBuilderException(
            "dxworkflow.json/source workflow spec contains no 'version' field, but it is required to build a global workflow")
    if not dxpy.executable_builder.GLOBAL_EXEC_VERSION_RE.match(json_spec['version']):
        logger.warn('"version" {} should be semver compliant (e.g. of the form X.Y.Z)'.format(json_spec['version']))

    if 'details' in json_spec:
        if not isinstance(json_spec['details'], dict):
            raise WorkflowBuilderException(
                'The field "details" must be a dictionary')

    if 'regionalOptions' in json_spec:
        if not (isinstance(json_spec['regionalOptions'], dict)
                and json_spec['regionalOptions']
                and all([isinstance(i, dict) for i in json_spec['regionalOptions'].values()])):
            raise WorkflowBuilderException(
                'The field "regionalOptions" must be a non-empty dictionary whose values are dictionaries')



def _get_validated_json(json_spec, args):
    """
    Validates workflow spec and returns the json that can be sent with the
    /workflow/new API or /globalworkflow/new request.
    """
    if not json_spec:
        return
    if not args:
        return

    validated_spec = copy.deepcopy(json_spec)

    # print ignored keys if present in json_spec
    unsupported_keys = _get_unsupported_keys(validated_spec.keys(), SUPPORTED_KEYS)
    if len(unsupported_keys) > 0:
        logger.warn(
            "Warning: the following root level fields are not supported and will be ignored: {}"
                .format(", ".join(unsupported_keys)))


    if 'stages' in validated_spec:
        validated_spec['stages'] = _get_validated_stages(validated_spec['stages'])

    if 'name' in validated_spec and not args._from:
        dir_name = os.path.basename(os.path.normpath(args.src_dir))
        if dir_name != validated_spec['name']:
            logger.warn(
                'workflow name "{}" does not match containing directory "{}"'.format(validated_spec['name'], dir_name))

    if 'ignoreReuse' in validated_spec:
        validate_ignore_reuse(validated_spec['stages'], validated_spec['ignoreReuse'])

    validated_documentation_fields = _get_validated_json_for_build_or_update(validated_spec, args)
    validated_spec.update(validated_documentation_fields)

    # Project-based workflow specific validation
    if args.mode == 'workflow':
        validated = _validate_json_for_regular_workflow(json_spec, args)
        validated_spec.update(validated)

    # Global workflow specific validation
    if args.mode == 'globalworkflow':
        _validate_json_for_global_workflow(validated_spec, args)

    return validated_spec


def _get_validated_json_for_build_or_update(json_spec, args):
    """
    Validates those fields that can be used when either building
    a new version (of a local, project-based workflow) or updating
    an existing version (of a global workflow).
    """
    validated = copy.deepcopy(json_spec)

    if not args._from:
        dxpy.executable_builder.inline_documentation_files(validated, args.src_dir)

    if 'title' not in json_spec:
        logger.warn("dxworkflow.json/source workflow spec is missing a title, please add one in the 'title' field")

    if 'summary' not in json_spec:
        logger.warn("dxworkflow.json/source workflow spec is missing a summary, please add one in the 'summary' field")
    else:
        if json_spec['summary'].endswith('.'):
            logger.warn("summary {} should be a short phrase not ending in a period".format(json_spec['summary'],))

    return validated


def _assert_executable_regions_match(workflow_enabled_regions, workflow_spec):
    """
    Check if the dependent apps/applets/subworkflows in the workflow are enabled in requested regions
    Returns the subset of requested regions where all the dependent executables are enabled
    If the workflow contains any applets, then the workflow can be currently enabled
    in only one region - the region in which the applets are stored.
    """
    if not workflow_enabled_regions: # empty set
        return workflow_enabled_regions
    
    # get executable from all stages and sort them in the order app/applet/globalworkflow/workflow
    executables = sorted([i.get("executable") for i in workflow_spec.get("stages")])

    for exect in executables:
        if exect.startswith("applet-"):
            applet_project = dxpy.DXApplet(exect).project
            applet_region = dxpy.DXProject(applet_project).region
            if {applet_region} != workflow_enabled_regions:                
                raise WorkflowBuilderException("The applet {} is not available in all requested region(s) {}"
                                               .format(exect, ','.join(workflow_enabled_regions)))

        elif exect.startswith("app-"):
            app_regional_options = dxpy.api.app_describe(exect, input_params={"fields": {"regionalOptions": True}})
            app_regions = set(app_regional_options['regionalOptions'].keys())
            if not workflow_enabled_regions.issubset(app_regions):
                additional_workflow_regions = workflow_enabled_regions.difference(app_regions)
                mesg = "The app {} is enabled in regions {} while the global workflow in {}.".format(
                    exect, ", ".join(app_regions), ", ".join(workflow_enabled_regions))
                mesg += " The workflow will not be able to run in {}.".format(", ".join(additional_workflow_regions))
                mesg += " If you are a developer of the app, you can enable the app in {} to run the workflow in that region(s).".format(
                    ", ".join(additional_workflow_regions))
                logger.warn(mesg)
                workflow_enabled_regions.intersection_update(app_regions)

        elif exect.startswith("workflow-"):
             # We recurse to check the regions of the executables of the inner workflow
            inner_workflow_spec = dxpy.api.workflow_describe(exect)
            workflow_enabled_regions = _assert_executable_regions_match(workflow_enabled_regions, inner_workflow_spec)

        elif exect.startswith("globalworkflow-"):
            raise WorkflowBuilderException("Building a global workflow with nested global workflows is not yet supported")

    return workflow_enabled_regions

def _build_regular_workflow(json_spec, keep_open=False):
    """
    Precondition: json_spec must be validated
    """
    workflow_id = dxpy.api.workflow_new(json_spec)["id"]
    if not keep_open:
        dxpy.api.workflow_close(workflow_id)
    return workflow_id


def _get_validated_enabled_regions(json_spec, args):
    """
    Returns a set of regions (region names) in which the global workflow
    should be enabled. 
    
    1. validates and synchronizes the regions passed via CLI argument and in the regionalOptions field.
    2. checks if these regions are included in the permitted regions of the bill_to
    3. checks if the dependencies in all the stages are enabled in these regions
    """
    # Determine in which regions the global workflow are requested to be available
    enabled_regions = dxpy.executable_builder.get_enabled_regions('globalworkflow',
                                                                  json_spec,
                                                                  args.region,
                                                                  WorkflowBuilderException)
    if not enabled_regions:
        enabled_regions = []
        if not dxpy.PROJECT_CONTEXT_ID:
            msg = "A context project must be selected to enable a workflow in the project's region."
            msg += " You can use 'dx select' to select a project. Otherwise you can use --region option"
            msg += " to select a region in which the workflow should be enabled"
            raise(WorkflowBuilderException(msg))
        current_selected_region = dxpy.api.project_describe(dxpy.PROJECT_CONTEXT_ID,
                                           input_params={"fields": {"region": True}})["region"]

        enabled_regions.append(current_selected_region)

    # Get billable regions
    enabled_regions = set(enabled_regions)
    billable_regions = dxpy.executable_builder.get_permitted_regions(
        json_spec["billTo"], WorkflowBuilderException)
    if not enabled_regions.issubset(billable_regions):
        raise WorkflowBuilderException("The global workflow cannot be enabled in regions {}, which are not among the permittedRegions of the billTo."
                                       .format(",".join(enabled_regions.difference(billable_regions))))

    # Verify dependencies in all the stages are also enabled in these regions
    enabled_regions = _assert_executable_regions_match(enabled_regions, json_spec)
    
    if not enabled_regions:
        raise AssertionError("This workflow should be enabled in at least one region.")

    return enabled_regions


def _create_temporary_projects(enabled_regions, bill_to):
    """
    Creates a temporary project needed to build an underlying workflow
    for a global workflow. Returns a dictionary with region names as keys
    and project IDs as values

    The regions in which projects will be created can be:
    i. regions specified in the dxworkflow.json/source workflow spec "regionalOptions"
    ii. regions specified as the argument "--region" when calling "dx build"
    iii. current context project, if none of the above are set
    iv. the regions where dependent applets/apps/workflows are enabled
    """
    # Create one temp project in each region
    projects_by_region = {}  # Project IDs by region
    for region in enabled_regions:
        try:
            project_input = {"name": "Temporary build project for dx build global workflow",
                             "region": region,
                             "billTo": bill_to}
            temp_project = dxpy.api.project_new(project_input)["id"]
            projects_by_region[region] = temp_project
            logger.debug("Created temporary project {} to build in".format(temp_project))
        except:
            # Clean up any temp projects that might have been created
            if projects_by_region:
                dxpy.executable_builder.delete_temporary_projects(projects_by_region.values())
            err_exit()
    return projects_by_region


def _build_underlying_workflows(enabled_regions, json_spec, args):
    """
    Creates a workflow in a temporary project for each enabled region.
    Returns a tuple of dictionaries: workflow IDs by region and project IDs by region.
    The caller is responsible for destroying the projects if this method returns properly.
    """
    projects_by_region = _create_temporary_projects(enabled_regions, json_spec["billTo"])
    workflows_by_region = {}

    try:
        for region, project in projects_by_region.items():
            # Override workflow project ID and folder in workflow spec 
            # when building underlying workflow in temporary project
            json_spec['project'] = project
            json_spec['folder'] = '/'
            workflow_id = _build_regular_workflow(json_spec)
            logger.debug("Created workflow " + workflow_id + " successfully")
            workflows_by_region[region] = workflow_id
    except:
        # Clean up
        if projects_by_region:
            dxpy.executable_builder.delete_temporary_projects(projects_by_region.values())
        raise

    return workflows_by_region, projects_by_region


def _build_global_workflow(json_spec, enabled_regions, args):
    """
    Creates a workflow in a temporary project for each enabled region
    and builds a global workflow on the platform based on these workflows.
    """
    workflows_by_region, projects_by_region = {}, {}  # IDs by region
    try:
        # prepare "regionalOptions" field for the globalworkflow/new input
        existing_regional_options = json_spec.get('regionalOptions', {})

        if existing_regional_options:
            if set(existing_regional_options.keys()) != set(enabled_regions):
                raise WorkflowBuilderException("These enabled regions do have regional options specified \
                                           in the JSON spec or from --extra-args: {}",
                                           ",".join(set(enabled_regions).difference(set(existing_regional_options.keys()))))
                
            updated_regional_options= copy.deepcopy(existing_regional_options)
        else:
            updated_regional_options = dict.fromkeys(enabled_regions, {})
        
        workflows_by_region, projects_by_region = \
            _build_underlying_workflows(enabled_regions, json_spec, args)
        for region, workflow_id in workflows_by_region.items():
            updated_regional_options[region]["workflow"] = workflow_id

        # update existing regionalOptions with underlying workflow ids and ignore regions that are not enabled
        json_spec.update({'regionalOptions': updated_regional_options})

        # leave only fields that are actually used to build the workflow
        gwf_provided_keys = GLOBALWF_SUPPORTED_KEYS.intersection(set(json_spec.keys()))
        gwf_final_json = dict((k, v) for k, v in json_spec.items() if k in gwf_provided_keys)

        # we don't want to print the whole documentation to the screen so we'll remove these fields
        print_spec = copy.deepcopy(gwf_final_json)
        if "description" in gwf_final_json:
            del print_spec["description"]
        if "developerNotes" in gwf_final_json:
            del print_spec["developerNotes"]
        logger.info("Will create global workflow with spec: {}".format(json.dumps(print_spec)))

        # Create a new global workflow version on the platform
        global_workflow_id = dxpy.api.global_workflow_new(gwf_final_json)["id"]

        logger.info("Uploaded global workflow {n}/{v} successfully".format(n=gwf_final_json["name"],
                                                                           v=gwf_final_json["version"]))
        logger.info("You can publish this workflow with:")
        logger.info("  dx publish {n}/{v}".format(n=gwf_final_json["name"],
                                                  v=gwf_final_json["version"]))
    finally:
        # Clean up
        if projects_by_region:
            dxpy.executable_builder.delete_temporary_projects(projects_by_region.values())

    # Set any additional fields on the created workflow
    try:
        _set_categories_on_workflow(global_workflow_id, gwf_final_json.get("categories", []))
    except:
        logger.warn(
            "The workflow {n}/{v} was created but setting categories failed".format(n=gwf_final_json['name'],
                                                                                    v=gwf_final_json['version']))
        raise

    return global_workflow_id


def _update_global_workflow(json_spec, args, global_workflow_id):

    def skip_update():
        skip_update = False
        if non_empty_fields:
            update_message = "The global workflow {}/{} exists so we will update the following fields for this version: {}.".format(
                json_spec["name"], json_spec["version"], ", ".join(non_empty_fields))

            if args.confirm:
                if INTERACTIVE_CLI:
                    try:
                        print('***')
                        print(fill('INFO: ' + update_message))
                        print('***')
                        value = input('Confirm making these updates [y/N]: ')
                    except KeyboardInterrupt:
                        value = 'n'
                    if not value.lower().startswith('y'):
                        skip_update = True
                else:
                    # Default to NOT updating if operating without a TTY.
                    logger.warn(
                        'skipping requested change to update a global workflow version. Rerun "dx build" interactively or pass --yes to confirm this change.')
                    skip_update = True
            else:
                logger.info(update_message)
        else:
            skip_update = True
            logger.info("Nothing to update")
        return skip_update

    update_spec = dict((k, v) for k, v in json_spec.items() if k in UPDATABLE_GLOBALWF_FIELDS)
    validated_spec = _get_validated_json_for_build_or_update(update_spec, args)
    non_empty_fields = dict((k, v) for k, v in validated_spec.items() if v)

    if args.update and not skip_update():
        global_workflow_id = dxpy.api.global_workflow_update('globalworkflow-' + json_spec['name'],
                                                             alias=json_spec['version'],
                                                             input_params=non_empty_fields)['id']
    else:
        logger.info("Skipping making updates")
    return global_workflow_id


def _build_or_update_workflow(args, parser):
    """
    Creates or updates a workflow on the platform.
    Returns the workflow ID, or None if the workflow cannot be created.
    """
    try:
        if args.mode == 'workflow':
            json_spec = _fetch_spec_from_dxworkflowjson(args.src_dir, "dxworkflow.json", parser)
            json_spec.update(args.extra_args or {})
            json_spec = _get_validated_json(json_spec, args)
            workflow_id = _build_regular_workflow(json_spec, args.keep_open)
        elif args.mode == 'globalworkflow':
            if args._from:
                json_spec = args._from
            else:
                json_spec = _fetch_spec_from_dxworkflowjson(args.src_dir, "dxworkflow.json", parser)
            
            # Override version number in json_spec if --version is specified
            # version is required when building global workflow but is optional for workflow
            # so `dx build` requires --version to be specified when using the --from
            if args.version_override:
                json_spec["version"] = args.version_override

            json_spec.update(args.extra_args or {})
            json_spec = _get_validated_json(json_spec, args)
            
            # Check if the local or source workflow is compiled by dxCompiler that supported dependency annotation
            if json_spec.get("tags") and "dxCompiler" in json_spec["tags"]:
                _check_dxcompiler_version(json_spec)
                if not args.brief:
                    _notify_instance_type_selection(json_spec)
                    _notify_dependencies(json_spec)
            
            json_spec = _cleanup_empty_keys(json_spec)
            # Verify if the global workflow already exists and if the user has developer rights to it
            # If the global workflow name doesn't exist, the user is free to build it
            # If the name does exist two things can be done:
            # * either update the requested version, if this version already exists
            # * or create the version if it doesn't exist
            existing_workflow = dxpy.executable_builder.verify_developer_rights('globalworkflow-' + json_spec['name'])
            if existing_workflow and _version_exists(json_spec,
                                                     existing_workflow.name,
                                                     existing_workflow.version):
                workflow_id = _update_global_workflow(json_spec, args, existing_workflow.id)
            else:
                json_spec["billTo"] = dxpy.executable_builder.get_valid_bill_to(args.bill_to, WorkflowBuilderException)
                enabled_regions = _get_validated_enabled_regions(json_spec, args)
                workflow_id = _build_global_workflow(json_spec, enabled_regions, args)
        else:
            raise WorkflowBuilderException("Unrecognized workflow type: {}".format(args.mode))
    except dxpy.exceptions.DXAPIError as e:
        raise e
    return workflow_id


def _print_output(workflow_id, args):
    if args.json and args.mode == 'workflow':
        output = dxpy.api.workflow_describe(workflow_id)
    elif args.json and args.mode == 'globalworkflow':
        output = dxpy.api.global_workflow_describe(workflow_id)
    else:
        output = {'id': workflow_id}
    if output is not None:
        print(json.dumps(output))


def build(args, parser):
    """
    Validates workflow source directory and creates a new (global) workflow based on it.
    Raises: WorkflowBuilderException if the workflow cannot be created.
    """
    if args is None:
        raise Exception("Arguments not provided")

    try:
        process_extra_args(args)
        workflow_id = _build_or_update_workflow(args, parser)
        _print_output(workflow_id, args)
    except WorkflowBuilderException as e:
        print("Error: {}".format(e), file=sys.stderr)
        sys.exit(3)



# Copyright (C) 2014-2016 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

'''
This module handles download commands for the dx command-line client.
'''
from __future__ import print_function, unicode_literals, division, absolute_import

import collections
import os
import subprocess
import sys
import logging

import dxpy
from ..utils.resolver import (resolve_existing_path, get_first_pos_of_char, is_project_explicit,
                              object_exists_in_project, is_jbor_str)
from ..exceptions import err_exit
from . import try_call
from dxpy.utils.printing import (fill)
from dxpy.utils import pathmatch

# Check if a program (wget, curl, etc.) is on the path, and
# can be called.
def _which(program):
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    for path in os.environ["PATH"].split(os.pathsep):
        exe_file = os.path.join(path, program)
        if is_exe(exe_file):
            return exe_file
    return None

# Caluclate the md5 checkum for [filename], and raise
# an exception if the checksum is wrong.
def _verify(filename, md5digest):
    md5sum_exe = _which("md5sum")
    if md5sum_exe is None:
        err_exit("md5sum is not installed on this system")
    cmd = [md5sum_exe, "-b", filename]
    try:
        print("Calculating checksum")
        cmd_out = subprocess.check_output(cmd)
    except subprocess.CalledProcessError:
        err_exit("Failed to run md5sum: " + str(cmd))

    line = cmd_out.strip().split()
    if len(line) != 2:
        err_exit("md5sum returned weird results: " + str(line))
    actual_md5 = line[0]
    if actual_md5 != md5digest:
        err_exit("Checksum doesn't match " + actual_md5 + "  expected:" + md5digest)
    print("Checksum correct")


def download_one_file(project, file_desc, dest_filename, args):
    if not args.overwrite:
        if os.path.exists(dest_filename):
            err_exit(fill('Error: path "' + dest_filename + '" already exists but -f/--overwrite was not set'))

    if file_desc['class'] != 'file':
        print("Skipping non-file data object {name} ({id})".format(**file_desc), file=sys.stderr)
        return

    if file_desc['state'] != 'closed':
        print("Skipping file {name} ({id}) because it is not closed".format(**file_desc), file=sys.stderr)
        return

    try:
        show_progress = args.show_progress
    except AttributeError:
        show_progress = False


    try:
        symlink_max_tries = args.symlink_max_tries if vars(args).get('symlink_max_tries') is not None else 15
        dxpy.download_dxfile(
                            file_desc['id'],
                            dest_filename,
                            show_progress=show_progress,
                            project=project,
                            describe_output=file_desc,
                            symlink_max_tries=symlink_max_tries)
        return
    except:
        err_exit()

def do_debug(msg):
    logging.debug(msg)

# dest_filename = local file where downloaded file will go
# src_filename = name of parquet file or folder being downloaded from database
def download_one_database_file(project, database_desc, dest_filename, src_filename, file_status, args):
    do_debug("download.py#download_one_database_file - src_filename = {}".format(src_filename));
    if file_status is not None:
        do_debug("download.py#download_one_database_file - file_status = {}".format(file_status));

    if not args.overwrite:
        if os.path.exists(dest_filename):
            err_exit(fill('Error: path "' + dest_filename + '" already exists but -f/--overwrite was not set'))

    if database_desc['class'] != 'database':
        print("Skipping non-database data object {name} ({id})".format(**database_desc), file=sys.stderr)
        return

    try:
        show_progress = args.show_progress
    except AttributeError:
        show_progress = False

    try:
        dxpy.download_dxdatabasefile(
            database_desc['id'],
            dest_filename,
            src_filename,
            file_status,
            show_progress=show_progress,
            project=project,
            describe_output=database_desc)
    except:
        err_exit()


def _ensure_local_dir(d):
    if not os.path.isdir(d):
        if os.path.exists(d):
            err_exit(fill('Error: path "' + d + '" already exists and is not a directory'))
        os.makedirs(d)


def _is_glob(path):
    return get_first_pos_of_char('*', path) > -1 or get_first_pos_of_char('?', path) > -1


def _rel2abs(path, project):
    if path.startswith('/') or dxpy.WORKSPACE_ID != project:
        abs_path, strip_prefix = path, os.path.dirname(path.rstrip('/'))
    else:
        wd = dxpy.config.get('DX_CLI_WD', u'/')
        abs_path, strip_prefix = os.path.join(wd, path), os.path.dirname(os.path.join(wd, path).rstrip('/'))
    if len(abs_path) > 1:
        abs_path = abs_path.rstrip('/')
    return abs_path, strip_prefix


def _download_files(files, destdir, args, dest_filename=None):
    for project in files:
        for f in files[project]:
            file_desc = f['describe']
            dest = dest_filename or os.path.join(destdir, file_desc['name'].replace('/', '%2F'))
            download_one_file(project, file_desc, dest, args)


def _download_folders(folders, destdir, args):
    try:
        show_progress = args.show_progress
    except AttributeError:
        show_progress = False
    for project in folders:
        for folder, strip_prefix in folders[project]:
            if not args.recursive:
                err_exit('Error: "' + folder + '" is a folder but the -r/--recursive option was not given')
            assert(folder.startswith(strip_prefix))
            folder_destdir = os.path.join(destdir, folder[len(strip_prefix):].lstrip('/'))
            try:
                dxpy.download_folder(project, folder_destdir, folder=folder, overwrite=args.overwrite,
                                     show_progress=show_progress)
            except:
                err_exit()


# Main entry point.
def download(args):
    folders_to_get, files_to_get, count = collections.defaultdict(list), collections.defaultdict(list), 0
    foldernames, filenames = [], []
    for path in args.paths:
        # Attempt to resolve name. If --all is given or the path looks like a glob, download all matches.
        # Otherwise, the resolver will display a picker (or error out if there is no tty to display to).
        resolver_kwargs = {'allow_empty_string': False}
        if args.all or _is_glob(path):
            resolver_kwargs.update({'allow_mult': True, 'all_mult': True})

        # include "parts" and a few additional fields in the description so that
        # we don't have to call a separate describe method downstream
        resolver_kwargs.update({"describe": {"parts": True,
                                             "size": True,
                                             "drive": True,
                                             "md5": True,
                                             "perPartCheckSum": True
                                             }})

        project, folderpath, matching_files = try_call(resolve_existing_path, path, **resolver_kwargs)

        if matching_files is None:
            matching_files = []
        elif not isinstance(matching_files, list):
            matching_files = [matching_files]

        # TODO: this could also be returned as metadata by resolve_path since
        # resolve_path knows these things in some circumstances
        path_has_explicit_proj = is_project_explicit(path) or is_jbor_str(path)

        if is_jbor_str(path):
            assert len(matching_files) == 1
            project = matching_files[0]["describe"]["project"]
        matching_folders = []
        # project may be none if path is an ID and there is no project context
        if project is not None:
            colon_pos = get_first_pos_of_char(":", path)
            if colon_pos >= 0:
                path = path[colon_pos + 1:]
            abs_path, strip_prefix = _rel2abs(path, project)
            parent_folder = os.path.dirname(abs_path)
            folder_listing = dxpy.list_subfolders(project, parent_folder, recurse=False)
            matching_folders = pathmatch.filter(folder_listing, abs_path)
            if '/' in matching_folders and len(matching_folders) > 1:
                # The list of subfolders is {'/', '/A', '/B'}.
                # Remove '/', otherwise we will download everything twice.
                matching_folders.remove('/')

        if len(matching_files) == 0 and len(matching_folders) == 0:
            err_exit(fill('Error: {path} is neither a file nor a folder name'.format(path=path)))

        # If the user did not explicitly provide the project, don't pass any
        # project parameter to the API call but continue with the download.
        if not path_has_explicit_proj:
            project = dxpy.DXFile.NO_PROJECT_HINT

        # If the user explicitly provided the project and it doesn't contain
        # the files, don't allow the download.

        # For speed's sake, skip this check (i.e. one API call) if the user
        # passed the lightweight argument
        #
        # If length of matching_files is 0 then we're only downloading folders
        # so skip this logic since the files will be verified in the API call.
        if not args.lightweight \
            and len(matching_files) > 0 \
            and path_has_explicit_proj \
            and not any(object_exists_in_project(f['describe']['id'], project) for f in matching_files):
                err_exit(fill('Error: specified project does not contain specified file object'))

        files_to_get[project].extend(matching_files)
        folders_to_get[project].extend(((f, strip_prefix) for f in matching_folders))
        count += len(matching_files) + len(matching_folders)
        filenames.extend(f["describe"]["name"] for f in matching_files)
        foldernames.extend(f[len(strip_prefix):].lstrip('/') for f in matching_folders)

    if len(filenames) > 0 and len(foldernames) > 0:
        name_conflicts = set(filenames) & set(foldernames)
        if len(name_conflicts) > 0:
            msg = "Error: The following paths are both file and folder names, and " \
                  "cannot be downloaded to the same destination: "
            msg += ", ".join(sorted(name_conflicts))
            err_exit(fill(msg))

    if args.output is None:
        destdir, dest_filename = os.getcwd(), None
    elif count > 1:
        if not os.path.exists(args.output):
            err_exit(fill("Error: When downloading multiple objects, --output must be an existing directory"))
        destdir, dest_filename = args.output, None
    elif os.path.isdir(args.output):
        destdir, dest_filename = args.output, None
    elif args.output.endswith('/'):
        err_exit(fill("Error: {path} could not be found".format(path=args.output)))
    else:
        destdir, dest_filename = os.getcwd(), args.output

    _download_folders(folders_to_get, destdir, args)
    _download_files(files_to_get, destdir, args, dest_filename=dest_filename)



# Copyright (C) 2013-2016 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

'''
App Builder Library
+++++++++++++++++++

Contains utility methods useful for compiling and deploying applets and apps
onto the platform.

You can specify the destination project in the following ways (with the earlier
ones taking precedence):

* Supply the *project* argument to :func:`upload_resources()` or
  :func:`upload_applet()`.
* Supply the 'project' attribute in your ``dxapp.json``.
* Set the ``DX_WORKSPACE_ID`` environment variable (when running in a job context).

You can use the function :func:`get_destination_project` to determine
the effective destination project.

'''

from __future__ import print_function, unicode_literals, division, absolute_import

import os, sys, json, subprocess, tempfile, multiprocessing
import datetime
import gzip
import hashlib
import io
import tarfile
import stat

import dxpy
import dxpy.executable_builder
from . import logger
from .utils import merge
from .utils.printing import fill
from .cli import INTERACTIVE_CLI

NUM_CORES = multiprocessing.cpu_count()

DX_TOOLKIT_PKGS = ('dx-toolkit',)
DX_TOOLKIT_GIT_URLS = ("git@github.com:dnanexus/dx-toolkit.git",)


class AppBuilderException(Exception):
    """
    This exception is raised by the methods in this module when app or applet
    building fails.
    """
    pass

def _validate_applet_spec(applet_spec):
    if 'runSpec' not in applet_spec:
        raise AppBuilderException("Required field 'runSpec' not found in dxapp.json")

def _validate_app_spec(app_spec):
    pass

def _get_applet_spec(src_dir):
    applet_spec_file = os.path.join(src_dir, "dxapp.json")
    with open(applet_spec_file) as fh:
        applet_spec = json.load(fh)

    _validate_applet_spec(applet_spec)
    if 'project' not in applet_spec:
        applet_spec['project'] = dxpy.WORKSPACE_ID
    return applet_spec

def _get_app_spec(src_dir):
    app_spec_file = os.path.join(src_dir, "dxapp.json")
    with open(app_spec_file) as fh:
        app_spec = json.load(fh)

    _validate_app_spec(app_spec)
    return app_spec

def build(src_dir, parallel_build=True):
    """
    Runs any build scripts that are found in the specified directory.

    In particular, runs ``./configure`` if it exists, followed by ``make -jN``
    if it exists (building with as many parallel tasks as there are CPUs on the
    system).
    """
    # TODO: use Gentoo or deb buildsystem
    config_script = os.path.join(src_dir, "configure")
    if os.path.isfile(config_script) and os.access(config_script, os.X_OK):
        logger.debug("Running ./configure in {cwd}".format(cwd=os.path.abspath(src_dir)))
        try:
            subprocess.check_call([config_script])
        except subprocess.CalledProcessError as e:
            raise AppBuilderException("./configure in target directory failed with exit code %d" % (e.returncode,))
    if os.path.isfile(os.path.join(src_dir, "Makefile")) \
        or os.path.isfile(os.path.join(src_dir, "makefile")) \
        or os.path.isfile(os.path.join(src_dir, "GNUmakefile")):
        if parallel_build:
            make_shortcmd = "make -j%d" % (NUM_CORES,)
        else:
            make_shortcmd = "make"
        logger.debug("Building with {make} in {cwd}".format(make=make_shortcmd, cwd=os.path.abspath(src_dir)))
        try:
            make_cmd = ["make", "-C", src_dir]
            if parallel_build:
                make_cmd.append("-j" + str(NUM_CORES))
            subprocess.check_call(make_cmd)
        except subprocess.CalledProcessError as e:
            raise AppBuilderException("%s in target directory failed with exit code %d" % (make_shortcmd, e.returncode))

def get_destination_project(src_dir, project=None):
    """
    :returns: Project ID where applet specified by src_dir would be written
    :rtype: str

    Returns the project ID where the applet specified in *src_dir* (or
    its associated resource bundles) would be written. This returns the
    same project that would be used by :func:`upload_resources()` and
    :func:`upload_applet()`, given the same *src_dir* and *project*
    parameters.
    """
    if project is not None:
        return project
    return _get_applet_spec(src_dir)['project']

def is_link_local(link_target):
    """
    :param link_target: The target of a symbolic link, as given by os.readlink()
    :type link_target: string
    :returns: A boolean indicating the link is local to the current directory.
              This is defined to mean that os.path.isabs(link_target) == False
              and the link NEVER references the parent directory, so
              "./foo/../../curdir/foo" would return False.
    :rtype: boolean
    """
    is_local=(not os.path.isabs(link_target))

    if is_local:
        # make sure that the path NEVER extends outside the resources directory!
        d,l = os.path.split(link_target)
        link_parts = []
        while l:
            link_parts.append(l)
            d,l = os.path.split(d)
        curr_path = os.sep

        for p in reversed(link_parts):
            is_local = (is_local and not (curr_path == os.sep and p == os.pardir) )
            curr_path = os.path.abspath(os.path.join(curr_path, p))

    return is_local

def _fix_perms(perm_obj):
    """
    :param perm_obj: A permissions object, as given by os.stat()
    :type perm_obj: integer
    :returns: A permissions object that is the result of "chmod a+rX" on the
              given permission object.  This is defined to be the permission object
              bitwise or-ed with all stat.S_IR*, and if the stat.S_IXUSR bit is
              set, then the permission object should also be returned bitwise or-ed
              with stat.S_IX* (stat.S_IXUSR not included because it would be redundant).
    :rtype: integer
    """
    ret_perm = perm_obj | stat.S_IROTH | stat.S_IRGRP | stat.S_IRUSR
    if ret_perm & stat.S_IXUSR:
        ret_perm = ret_perm | stat.S_IXGRP | stat.S_IXOTH

    return ret_perm

def _fix_perm_filter(tar_obj):
    """
    :param tar_obj: A TarInfo object to be added to a tar file
    :tpye tar_obj: tarfile.TarInfo
    :returns: A TarInfo object with permissions changed (a+rX)
    :rtype: tarfile.TarInfo
    """
    tar_obj.mode = _fix_perms(tar_obj.mode)
    return tar_obj


def upload_resources(src_dir, project=None, folder='/', ensure_upload=False, force_symlinks=False, brief=False, resources_dir=None, worker_resources_subpath=""):
    """
    :param ensure_upload: If True, will bypass checksum of resources directory
                          and upload resources bundle unconditionally;
                          will NOT be able to reuse this bundle in future builds.
                          Else if False, will compute checksum and upload bundle
                          if checksum is different from a previously uploaded
                          bundle's checksum.
    :type ensure_upload: boolean
    :param force_symlinks: If true, will bypass the attempt to dereference any
                           non-local symlinks and will unconditionally include
                           the link as-is.  Note that this will almost certainly
                           result in a broken link within the resource directory
                           unless you really know what you're doing.
    :type force_symlinks: boolean
    :param resources_dir: Directory with resources to be archived and uploaded. If not given, uses `resources/`.
    :type resources_dir: str
    :param worker_resources_subpath: Path that will be prepended to the default directory where files are extracted on the worker.
                                     Default is empty string, therefore files would be extracted directly to the root folder.
                                     Example: If "home/dnanexus" is given, files will be extracted into /home/dnanexus.
    :type worker_resources_subpath: str
    :returns: A list (possibly empty) of references to the generated archive(s)
    :rtype: list

    If resources_dir exists, archives and uploads the contents of the resources_dir
    (usually ``resources/``) subdirectory of *src_dir* to a new remote file
    object, and returns a list describing a single bundled dependency in
    the form expected by the ``bundledDepends`` field of a run
    specification. Returns an empty list, if no archive was created.
    """
    if not resources_dir:
        resources_dir = os.path.join(src_dir, "resources")

    applet_spec = _get_applet_spec(src_dir)

    if project is None:
        dest_project = applet_spec['project']
    else:
        dest_project = project
        applet_spec['project'] = project

    if os.path.exists(resources_dir) and len(os.listdir(resources_dir)) > 0:
        target_folder = applet_spec['folder'] if 'folder' in applet_spec else folder

        # While creating the resource bundle, optimistically look for a
        # resource bundle with the same contents, and reuse it if possible.
        # The resource bundle carries a property 'resource_bundle_checksum'
        # that indicates the checksum; the way in which the checksum is
        # computed is given below.   If the checksum matches  (and
        # ensure_upload is False), then we will use the existing file,
        # otherwise, we will compress and upload the tarball.


        # The input to the SHA1 contains entries of the form (whitespace
        # only included here for readability):
        #
        # / \0 MODE \0 MTIME \0
        # /foo \0 MODE \0 MTIME \0
        # ...
        #
        # where there is one entry for each directory or file (order is
        # specified below), followed by a numeric representation of the
        # mode, and the mtime in milliseconds since the epoch.
        #
        # Note when looking at a link, if the link is to be dereferenced,
        # the mtime and mode used are that of the target (using os.stat())
        # If the link is to be kept as a link, the mtime and mode are those
        # of the link itself (using os.lstat())

        with tempfile.NamedTemporaryFile(suffix=".tar") as tar_tmp_fh:

            output_sha1 = hashlib.sha1()
            tar_fh = tarfile.open(fileobj=tar_tmp_fh, mode='w')

            for dirname, subdirs, files in os.walk(resources_dir):
                if not dirname.startswith(resources_dir):
                    raise AssertionError('Expected %r to start with root directory %r' % (dirname, resources_dir))

                # Add an entry for the directory itself
                relative_dirname = dirname[len(resources_dir):]
                dir_stat = os.lstat(dirname)
                if not relative_dirname.startswith('/'):
                    relative_dirname = '/' + relative_dirname

                fields = [relative_dirname, str(_fix_perms(dir_stat.st_mode)), str(int(dir_stat.st_mtime * 1000))]
                output_sha1.update(b''.join(s.encode('utf-8') + b'\0' for s in fields))

                # add an entry in the tar file for the current directory, but
                # do not recurse!
                tar_fh.add(dirname, arcname=worker_resources_subpath + relative_dirname, recursive=False, filter=_fix_perm_filter)
                # Canonicalize the order of subdirectories; this is the order in
                # which they will be visited by os.walk
                subdirs.sort()

                # check the subdirectories for symlinks.  We should throw an error
                # if there are any links that point outside of the directory (unless
                # --force-symlinks is given).  If a link is pointing internal to
                # the directory (or --force-symlinks is given), we should add it
                # as a file.
                for subdir_name in subdirs:
                    dir_path = os.path.join(dirname, subdir_name)

                    # If we do have a symlink,
                    if os.path.islink(dir_path):
                        # Let's get the pointed-to path to ensure that it is
                        # still in the directory
                        link_target = os.readlink(dir_path)

                        # If this is a local link, add it to the list of files (case 1)
                        # else raise an error
                        if force_symlinks or is_link_local(link_target):
                            files.append(subdir_name)
                        else:
                            raise AppBuilderException("Cannot include symlinks to directories outside of the resource directory.  '%s' points to directory '%s'" % (dir_path, os.path.realpath(dir_path)))


                # Canonicalize the order of files so that we compute the
                # checksum in a consistent order
                for filename in sorted(files):
                    deref_link = False

                    relative_filename = os.path.join(relative_dirname, filename)
                    true_filename = os.path.join(dirname, filename)

                    file_stat = os.lstat(true_filename)
                    # check for a link here, please!
                    if os.path.islink(true_filename):

                        # Get the pointed-to path
                        link_target = os.readlink(true_filename)

                        if not (force_symlinks or is_link_local(link_target)):
                            # if we are pointing outside of the directory, then:
                            # try to get the true stat of the file and make sure
                            # to dereference the link!
                            try:
                                file_stat = os.stat(os.path.join(dirname, link_target))
                                deref_link = True
                            except OSError:
                                # uh-oh! looks like we have a broken link!
                                # since this is guaranteed to cause problems (and
                                # we know we're not forcing symlinks here), we
                                # should throw an error
                                raise AppBuilderException("Broken symlink: Link '%s' points to '%s', which does not exist" % (true_filename, os.path.realpath(true_filename)) )


                    fields = [relative_filename, str(_fix_perms(file_stat.st_mode)), str(int(file_stat.st_mtime * 1000))]
                    output_sha1.update(b''.join(s.encode('utf-8') + b'\0' for s in fields))

                    # If we are to dereference, use the target fn
                    if deref_link:
                        true_filename = os.path.realpath(true_filename)
                    tar_fh.add(true_filename, arcname=worker_resources_subpath + relative_filename, filter=_fix_perm_filter)
                # end for filename in sorted(files)

            # end for dirname, subdirs, files in os.walk(resources_dir):

            # at this point, the tar is complete, so close the tar_fh
            tar_fh.close()

            # Optimistically look for a resource bundle with the same
            # contents, and reuse it if possible. The resource bundle
            # carries a property 'resource_bundle_checksum' that indicates
            # the checksum; the way in which the checksum is computed is
            # given in the documentation of _directory_checksum.

            if ensure_upload:
                properties_dict = {}
                existing_resources = False
            else:
                directory_checksum = output_sha1.hexdigest()
                properties_dict = dict(resource_bundle_checksum=directory_checksum)
                existing_resources = dxpy.find_one_data_object(
                    project=dest_project,
                    folder=target_folder,
                    properties=dict(resource_bundle_checksum=directory_checksum),
                    visibility='either',
                    zero_ok=True,
                    state='closed',
                    return_handler=True
                )

            if existing_resources:
                if not brief:
                    logger.info("Found existing resource bundle that matches local resources directory: " +
                                existing_resources.get_id())

                dx_resource_archive = existing_resources
            else:

                logger.debug("Uploading in " + src_dir)
                # We need to compress the tar that we've created


                targz_fh = tempfile.NamedTemporaryFile(suffix=".tar.gz", delete=False)

                # compress the file by reading the tar file and passing
                # it though a GzipFile object, writing the given
                # block size (by default 8192 bytes) at a time
                targz_gzf = gzip.GzipFile(fileobj=targz_fh, mode='wb')
                tar_tmp_fh.seek(0)
                dat = tar_tmp_fh.read(io.DEFAULT_BUFFER_SIZE)
                while dat:
                    targz_gzf.write(dat)
                    dat = tar_tmp_fh.read(io.DEFAULT_BUFFER_SIZE)

                targz_gzf.flush()
                targz_gzf.close()
                targz_fh.close()

                if 'folder' in applet_spec:
                    try:
                        dxpy.get_handler(dest_project).new_folder(applet_spec['folder'], parents=True)
                    except dxpy.exceptions.DXAPIError:
                        pass # TODO: make this better

                dx_resource_archive = dxpy.upload_local_file(
                    targz_fh.name,
                    wait_on_close=True,
                    project=dest_project,
                    folder=target_folder,
                    hidden=True,
                    properties=properties_dict
                )

                os.unlink(targz_fh.name)

                # end compressed file creation and upload

            archive_link = dxpy.dxlink(dx_resource_archive.get_id())

        # end tempfile.NamedTemporaryFile(suffix=".tar") as tar_fh

        return [{'name': 'resources.tar.gz', 'id': archive_link}]
    else:
        return []


def upload_applet(src_dir, uploaded_resources, check_name_collisions=True, overwrite=False, archive=False,
                  project=None, override_folder=None, override_name=None,
                  dry_run=False, brief=False, **kwargs):
    """
    Creates a new applet object.

    :param project: ID of container in which to create the applet.
    :type project: str, or None to use whatever is specified in dxapp.json
    :param override_folder: folder name for the resulting applet which, if specified, overrides that given in dxapp.json
    :type override_folder: str
    :param override_name: name for the resulting applet which, if specified, overrides that given in dxapp.json
    :type override_name: str

    """
    applet_spec = _get_applet_spec(src_dir)

    if project is None:
        dest_project = applet_spec['project']
    else:
        dest_project = project
        applet_spec['project'] = project

    if 'name' not in applet_spec:
        try:
            applet_spec['name'] = os.path.basename(os.path.abspath(src_dir))
        except:
            raise AppBuilderException("Could not determine applet name from the specification (dxapp.json) or from the name of the working directory (%r)" % (src_dir,))

    if override_folder:
        applet_spec['folder'] = override_folder
    if 'folder' not in applet_spec:
        applet_spec['folder'] = '/'

    if override_name:
        applet_spec['name'] = override_name

    if 'dxapi' not in applet_spec:
        applet_spec['dxapi'] = dxpy.API_VERSION

    applets_to_overwrite = []
    archived_applet = None
    if check_name_collisions and not dry_run:
        destination_path = applet_spec['folder'] + ('/' if not applet_spec['folder'].endswith('/') else '') + applet_spec['name']
        logger.debug("Checking for existing applet at " + destination_path)
        for result in dxpy.find_data_objects(classname="applet", name=applet_spec["name"], folder=applet_spec['folder'], project=dest_project, recurse=False):
            if overwrite:
                # Don't remove the old applet until after the new one
                # has been created. This avoids a race condition where
                # we remove the old applet, but that causes garbage
                # collection of the bundled resources that will be
                # shared with the new applet
                applets_to_overwrite.append(result['id'])
            elif archive:
                logger.debug("Archiving applet %s" % (result['id']))
                proj = dxpy.DXProject(dest_project)
                archive_folder = '/.Applet_archive'
                try:
                    proj.list_folder(archive_folder)
                except dxpy.DXAPIError:
                    proj.new_folder(archive_folder)

                proj.move(objects=[result['id']], destination=archive_folder)
                archived_applet = dxpy.DXApplet(result['id'], project=dest_project)
                now = datetime.datetime.fromtimestamp(archived_applet.created/1000).ctime()
                new_name = archived_applet.name + " ({d})".format(d=now)
                archived_applet.rename(new_name)
                if not brief:
                    logger.info("Archived applet %s to %s:\"%s/%s\"" % (result['id'], dest_project, archive_folder, new_name))
            else:
                raise AppBuilderException("An applet already exists at %s (id %s) and the --overwrite (-f) or --archive (-a) options were not given" % (destination_path, result['id']))

    # -----
    # Override various fields from the pristine dxapp.json
    # TODO: should merge extra args into app spec before processing
    #       see line https://github.com/dnanexus/dx-toolkit/blob/f2ebe53a9e49b3213f8252e7ac1cd35d99913333/src/python/dxpy/app_builder.py#L616
    # Carry region-specific values from regionalOptions into the main
    # runSpec
    applet_spec["runSpec"].setdefault("bundledDepends", [])
    applet_spec["runSpec"].setdefault("assetDepends", [])
    if not dry_run:
        region = dxpy.api.project_describe(dest_project, input_params={"fields": {"region": True}})["region"]

        # if regionalOptions contain at least one region, they must include
        # the region of the target project
        if len(applet_spec.get('regionalOptions', {})) != 0 and region not in applet_spec.get('regionalOptions', {}):
            err_mesg = "destination project is in region {} but \"regionalOptions\" do not contain this region. ".format(region)
            err_mesg += "Please, update your \"regionalOptions\" specification"
            raise AppBuilderException(err_mesg)

        regional_options = applet_spec.get('regionalOptions', {}).get(region, {})

        # We checked earlier that if region-specific values for the
        # fields below are given, the same fields are not also specified
        # in the top-level runSpec. So the operations below should not
        # result in any user-supplied settings being clobbered.

        if 'systemRequirements' in regional_options:
            applet_spec["runSpec"]["systemRequirements"] = regional_options['systemRequirements']

        if 'bundledDepends' in regional_options:
            applet_spec["runSpec"]["bundledDepends"].extend(regional_options["bundledDepends"])
        if 'assetDepends' in regional_options:
            applet_spec["runSpec"]["assetDepends"].extend(regional_options["assetDepends"])

    # Inline Readme.md and Readme.developer.md
    dxpy.executable_builder.inline_documentation_files(applet_spec, src_dir)

    # Inline the code of the program
    if "file" in applet_spec["runSpec"]:
        # Put it into runSpec.code instead
        with open(os.path.join(src_dir, applet_spec["runSpec"]["file"])) as code_fh:
            applet_spec["runSpec"]["code"] = code_fh.read()
            del applet_spec["runSpec"]["file"]

    # If this is applet requires a cluster, inline any bootstrapScript code that may be provided.
    # bootstrapScript is an *optional* clusterSpec parameter.
    # NOTE: assumes bootstrapScript is always provided as a filename
    if "systemRequirements" in applet_spec["runSpec"]:
        sys_reqs = applet_spec["runSpec"]["systemRequirements"]
        for entry_point in sys_reqs:
            try:
                bootstrap_script = os.path.join(src_dir, sys_reqs[entry_point]["clusterSpec"]["bootstrapScript"])
                with open(bootstrap_script) as code_fh:
                    sys_reqs[entry_point]["clusterSpec"]["bootstrapScript"] = code_fh.read()
            except KeyError:
                # either no "clusterSpec" or no "bootstrapScript" within "clusterSpec"
                continue
            except IOError:
                raise AppBuilderException("The clusterSpec \"bootstrapScript\" could not be read.")

    # Attach bundled resources to the app
    if uploaded_resources is not None:
        applet_spec["runSpec"]["bundledDepends"].extend(uploaded_resources)

    # Validate and process assetDepends
    asset_depends = applet_spec["runSpec"]["assetDepends"]
    if type(asset_depends) is not list or any(type(dep) is not dict for dep in asset_depends):
        raise AppBuilderException("Expected runSpec.assetDepends to be an array of objects")
    for asset in asset_depends:
        asset_project = asset.get("project", None)
        asset_folder = asset.get("folder", '/')
        asset_stages = asset.get("stages", None)
        if "id" in asset:
            asset_record = dxpy.DXRecord(asset["id"]).describe(fields={'details'}, default_fields=True)
        elif "name" in asset and asset_project is not None and "version" in asset:
            try:
                asset_record = dxpy.find_one_data_object(zero_ok=True, classname="record", typename="AssetBundle",
                                                         name=asset["name"], properties=dict(version=asset["version"]),
                                                         project=asset_project, folder=asset_folder, recurse=False,
                                                         describe={"defaultFields": True, "fields": {"details": True}},
                                                         state="closed", more_ok=False)
            except dxpy.exceptions.DXSearchError:
                msg = "Found more than one asset record that matches: name={0}, folder={1} in project={2}."
                raise AppBuilderException(msg.format(asset["name"], asset_folder, asset_project))
        else:
            raise AppBuilderException("Each runSpec.assetDepends element must have either {'id'} or "
                                      "{'name', 'project' and 'version'} field(s).")

        if asset_record:
            if "id" in asset:
                asset_details = asset_record["details"]
            else:
                asset_details = asset_record["describe"]["details"]
            if "archiveFileId" in asset_details:
                archive_file_id = asset_details["archiveFileId"]
            else:
                raise AppBuilderException("The required field 'archiveFileId' was not found in "
                                          "the details of the asset bundle %s " % asset_record["id"])
            archive_file_name = dxpy.DXFile(archive_file_id).describe()["name"]
            bundle_depends = {
                "name": archive_file_name,
                "id": archive_file_id
            }
            if asset_stages:
                bundle_depends["stages"] = asset_stages
            applet_spec["runSpec"]["bundledDepends"].append(bundle_depends)
            # If the file is not found in the applet destination project, clone it from the asset project
            if (not dry_run and
                    dxpy.DXRecord(dxid=asset_record["id"], project=dest_project).describe()["project"] != dest_project):
                dxpy.DXRecord(asset_record["id"], project=asset_record["project"]).clone(dest_project)
        else:
            raise AppBuilderException("No asset bundle was found that matched the specification %s"
                                      % (json.dumps(asset)))

    merge(applet_spec, kwargs)

    # -----
    # Now actually create the applet

    if dry_run:
        print("Would create the following applet:")
        print(json.dumps(applet_spec, indent=2))
        print("*** DRY-RUN-- no applet was created ***")
        return None, None

    if applet_spec.get("categories", []):
        if "tags" not in applet_spec:
            applet_spec["tags"] = []
        applet_spec["tags"] = list(set(applet_spec["tags"]) | set(applet_spec["categories"]))

    applet_id = dxpy.api.applet_new(applet_spec)["id"]

    if archived_applet:
        archived_applet.set_properties({'replacedWith': applet_id})

    # Now it is permissible to delete the old applet(s), if any
    if applets_to_overwrite:
        if not brief:
            logger.info("Deleting applet(s) %s" % (','.join(applets_to_overwrite)))
        dxpy.DXProject(dest_project).remove_objects(applets_to_overwrite)

    return applet_id, applet_spec


def _create_or_update_version(app_name, version, app_spec, try_update=True):
    """
    Creates a new version of the app. Returns an app_id, or None if the app has
    already been created and published.
    """
    # This has a race condition since the app could have been created or
    # published since we last looked.
    try:
        app_id = dxpy.api.app_new(app_spec)["id"]
        return app_id
    except dxpy.exceptions.DXAPIError as e:
        # TODO: detect this error more reliably
        if e.name == 'InvalidInput' and e.msg == 'Specified name and version conflict with an existing alias':
            print('App %s/%s already exists' % (app_spec["name"], version), file=sys.stderr)
            # The version number was already taken, so app/new doesn't work.
            # However, maybe it hasn't been published yet, so we might be able
            # to app-xxxx/update it.
            app_describe = dxpy.api.app_describe("app-" + app_name, alias=version)
            if app_describe.get("published", 0) > 0:
                return None
            return _update_version(app_name, version, app_spec, try_update=try_update)
        raise e

def _update_version(app_name, version, app_spec, try_update=True):
    """
    Updates a version of the app in place. Returns an app_id, or None if the
    app has already been published.
    """
    if not try_update:
        return None
    try:
        app_id = dxpy.api.app_update("app-" + app_name, version, app_spec)["id"]
        return app_id
    except dxpy.exceptions.DXAPIError as e:
        if e.name == 'InvalidState':
            print('App %s/%s has already been published' % (app_spec["name"], version), file=sys.stderr)
            return None
        raise e


def create_app_multi_region(regional_options, app_name, src_dir, publish=False, set_default=False, billTo=None,
                            try_versions=None, try_update=True, confirm=True, inherited_metadata={}, brief=False):
    """
    Creates a new app object from the specified applet(s).

    :param regional_options: Region-specific options for the app. See
        https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-new
        for details; this should contain keys for each region the app is
        to be enabled in, and for the values, a dict containing (at
        minimum) a key "applet" whose value is an applet ID for that
        region.
    :type regional_options: dict
    """
    return _create_app(dict(regionalOptions=regional_options), app_name, src_dir, publish=publish,
                       set_default=set_default, billTo=billTo, try_versions=try_versions, try_update=try_update,
                       confirm=confirm, inherited_metadata=inherited_metadata, brief=brief)


def create_app(applet_id, applet_name, src_dir, publish=False, set_default=False, billTo=None, try_versions=None,
               try_update=True, confirm=True, regional_options=None):
    """
    Creates a new app object from the specified applet.

    .. deprecated:: 0.204.0
       Use :func:`create_app_multi_region()` instead.

    """
    # In this case we don't know the region of the applet, so we use the
    # legacy API {"applet": applet_id} without specifying a region
    # specifically.
    return _create_app(dict(applet=applet_id), applet_name, src_dir, publish=publish, set_default=set_default,
                       billTo=billTo, try_versions=try_versions, try_update=try_update, confirm=confirm)


def _create_app(applet_or_regional_options, app_name, src_dir, publish=False, set_default=False, billTo=None,
                try_versions=None, try_update=True, confirm=True, inherited_metadata={}, brief=False):

    if src_dir:
        app_spec = _get_app_spec(src_dir)
        if not brief:
            logger.info("Will create app with spec: %s" % (json.dumps(app_spec),))
        # Inline Readme.md and Readme.developer.md
        dxpy.executable_builder.inline_documentation_files(app_spec, src_dir)
    else:
        app_spec = inherited_metadata

    app_spec.update(applet_or_regional_options, name=app_name)

    if billTo:
        app_spec["billTo"] = billTo
    if not try_versions:
        try_versions = [app_spec["version"]]

    for version in try_versions:
        logger.debug("Attempting to create version %s..." % (version,))
        app_spec['version'] = version

        app_describe = None
        try:
            # 404, which is rather likely in this app_describe request
            # (the purpose of the request is to find out whether the
            # version of interest exists), would ordinarily cause this
            # request to be retried multiple times, introducing a
            # substantial delay. So we disable retrying here for this
            # request.
            app_describe = dxpy.api.app_describe("app-" + app_spec["name"], alias=version, always_retry=False)
        except dxpy.exceptions.DXAPIError as e:
            if e.name == 'ResourceNotFound':
                pass
            else:
                raise e
        # Now app_describe is None if the app didn't exist, OR it contains the
        # app describe content.

        # The describe check does not eliminate race conditions since an app
        # may always have been created, or published, since we last looked at
        # it. So the describe that happens here is just to save time and avoid
        # unnecessary API calls, but we always have to be prepared to recover
        # from API errors.
        if app_describe is None:
            logger.debug('App %s/%s does not yet exist' % (app_spec["name"], version))
            app_id = _create_or_update_version(app_spec['name'], app_spec['version'], app_spec, try_update=try_update)
            if app_id is None:
                continue
            if not brief:
                logger.info("Created app " + app_id)
            # Success!
            break
        elif app_describe.get("published", 0) == 0:
            logger.debug('App %s/%s already exists and has not been published' % (app_spec["name"], version))
            app_id = _update_version(app_spec['name'], app_spec['version'], app_spec, try_update=try_update)
            if app_id is None:
                continue
            if not brief:
                logger.info("Updated existing app " + app_id)
            # Success!
            break
        else:
            logger.debug('App %s/%s already exists and has been published' % (app_spec["name"], version))
            # App has already been published. Give up on this version.
            continue
    else:
        # All versions requested failed
        if len(try_versions) != 1:
            tried_versions = 'any of the requested versions: ' + ', '.join(try_versions)
        else:
            tried_versions = 'the requested version: ' + try_versions[0]
        raise AppBuilderException('Could not create %s' % (tried_versions,))

    # Set categories appropriately.
    categories_to_set = app_spec.get("categories", [])
    existing_categories = dxpy.api.app_list_categories(app_id)['categories']
    categories_to_add = set(categories_to_set).difference(set(existing_categories))
    categories_to_remove = set(existing_categories).difference(set(categories_to_set))
    if categories_to_add:
        dxpy.api.app_add_categories(app_id, input_params={'categories': list(categories_to_add)})
    if categories_to_remove:
        dxpy.api.app_remove_categories(app_id, input_params={'categories': list(categories_to_remove)})

    # Set developers list appropriately, but only if provided.
    developers_to_set = app_spec.get("developers")
    if developers_to_set is not None:
        existing_developers = dxpy.api.app_list_developers(app_id)['developers']
        developers_to_add = set(developers_to_set) - set(existing_developers)
        developers_to_remove = set(existing_developers) - set(developers_to_set)

        skip_updating_developers = False
        if developers_to_add or developers_to_remove:
            parts = []
            if developers_to_add:
                parts.append('the following developers will be added: ' + ', '.join(sorted(developers_to_add)))
            if developers_to_remove:
                parts.append('the following developers will be removed: ' + ', '.join(sorted(developers_to_remove)))
            developer_change_message = '; and '.join(parts)
            if confirm:
                if INTERACTIVE_CLI:
                    try:
                        print('***')
                        print(fill('WARNING: ' + developer_change_message))
                        print('***')
                        value = input('Confirm updating developers list [y/N]: ')
                    except KeyboardInterrupt:
                        value = 'n'
                    if not value.lower().startswith('y'):
                        skip_updating_developers = True
                else:
                    # Default to NOT updating developers if operating
                    # without a TTY.
                    logger.warn('skipping requested change to the developer list. Rerun "dx build" interactively or pass --yes to confirm this change.')
                    skip_updating_developers = True
            else:
                logger.warn(developer_change_message)

        if not skip_updating_developers:
            if developers_to_add:
                dxpy.api.app_add_developers(app_id, input_params={'developers': list(developers_to_add)})
            if developers_to_remove:
                dxpy.api.app_remove_developers(app_id, input_params={'developers': list(developers_to_remove)})

    # Set authorizedUsers list appropriately, but only if provided.
    authorized_users_to_set = app_spec.get("authorizedUsers")
    existing_authorized_users = dxpy.api.app_list_authorized_users(app_id)['authorizedUsers']
    if authorized_users_to_set is not None:
        authorized_users_to_add = set(authorized_users_to_set) - set(existing_authorized_users)
        authorized_users_to_remove = set(existing_authorized_users) - set(authorized_users_to_set)

        skip_adding_public = False
        if 'PUBLIC' in authorized_users_to_add:
            acl_change_message = 'app-%s will be made public. Anyone will be able to view and run all published versions of this app.' % (app_spec['name'],)
            if confirm:
                if INTERACTIVE_CLI:
                    try:
                        print('***')
                        print(fill('WARNING: ' + acl_change_message))
                        print('***')
                        value = input('Confirm making this app public [y/N]: ')
                    except KeyboardInterrupt:
                        value = 'n'
                    if not value.lower().startswith('y'):
                        skip_adding_public = True
                else:
                    # Default to NOT adding PUBLIC if operating
                    # without a TTY.
                    logger.warn('skipping requested change to add PUBLIC to the authorized users list. Rerun "dx build" interactively or pass --yes to confirm this change.')
                    skip_adding_public = True
            else:
                logger.warn(acl_change_message)

        if skip_adding_public:
            authorized_users_to_add -= {'PUBLIC'}
        if authorized_users_to_add:
            dxpy.api.app_add_authorized_users(app_id, input_params={'authorizedUsers': list(authorized_users_to_add)})
        if skip_adding_public:
            logger.warn('the app was NOT made public as requested in the app spec. To make it so, run "dx add users app-%s PUBLIC".' % (app_spec["name"],))

        if authorized_users_to_remove:
            dxpy.api.app_remove_authorized_users(app_id, input_params={'authorizedUsers': list(authorized_users_to_remove)})

    elif not len(existing_authorized_users) and not brief:
        # Apps that had authorized users added by any other means will
        # not have this message printed.
        logger.warn('authorizedUsers is missing from the app spec. No one will be able to view or run the app except the app\'s developers.')

    if publish:
        dxpy.api.app_publish(app_id, input_params={'makeDefault': set_default})
    else:
        # If no versions of this app have ever been published, then
        # we'll set the "default" tag to point to the latest
        # (unpublished) version.
        no_published_versions = len(list(dxpy.find_apps(name=app_name, published=True, limit=1))) == 0
        if no_published_versions:
            dxpy.api.app_add_tags(app_id, input_params={'tags': ['default']})

    return app_id


def get_enabled_regions(app_spec, from_command_line):
    """Returns a list of the regions in which the app should be enabled.

    Also validates that app_spec['regionalOptions'], if supplied, is
    well-formed.

    :param app_spec: app specification
    :type app_spec: dict
    :param from_command_line: The regions specified on the command-line
      via --region
    :type from_command_line: list or None

    """
    enabled_regions = dxpy.executable_builder.get_enabled_regions('app', app_spec, from_command_line, AppBuilderException)

    if enabled_regions is not None and len(enabled_regions) == 0:
        raise AssertionError("This app should be enabled in at least one region")
    return enabled_regions



# Copyright (C) 2013-2016 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

'''
Exceptions for the :mod:`dxpy` package.
'''

from __future__ import print_function, unicode_literals, division, absolute_import

import sys
import json
import traceback
import errno
import socket
from urllib3.exceptions import HTTPError

import dxpy
import urllib3
import ssl

EXPECTED_ERR_EXIT_STATUS = 3

class DXError(Exception):
    '''Base class for exceptions in this package.'''

class DXAPIError(DXError):
    '''
    Exception for when the API server responds with a code that is not 200 (OK). See
    https://documentation.dnanexus.com/developer/api/protocols#errors for complete documentation of API errors,
    including those reflected by subclasses of this class.
    '''
    def __init__(self, content, code, timestamp="", req_id=""):
        self.name = content["error"]["type"]
        self.msg = content["error"]["message"]
        if "details" in content["error"]:
            self.details = content["error"]["details"]
        else:
            self.details = None
        self.code = code
        self.timestamp = timestamp
        self.req_id = req_id

    def error_message(self):
        "Returns a one-line description of the error."
        output = self.msg + ", code " + str(self.code)
        output += ". Request Time={}, Request ID={}".format(self.timestamp, self.req_id)
        if self.name != self.__class__.__name__:
            output = self.name + ": " + output
        return output

    def __str__(self):
        output = self.error_message()
        if self.details:
            output += "\nDetails: " + json.dumps(self.details, indent=4)
        return output

class MalformedJSON(DXAPIError):
    ''' Raised when the input could not be parsed as JSON. '''

class InvalidAuthentication(DXAPIError):
    ''' Raised when the provided OAuth2 token is invalid. '''

class PermissionDenied(DXAPIError):
    ''' Raised when the supplied credentials have insufficient permissions to perform this action. '''

class SpendingLimitExceeded(DXAPIError):
    ''' Raised when the spending limit has been reached for the account that would be billed for this action. '''

class ResourceNotFound(DXAPIError):
    ''' Raised when a specified entity or resource could not be found. '''

class InvalidInput(DXAPIError):
    ''' Raised when the input is syntactically correct (JSON), but semantically incorrect (for example, a JSON array
    is provided where a hash was required; or a required parameter was missing, etc.). '''

class InvalidState(DXAPIError):
    ''' Raised when the operation is not allowed at this object state. '''

class InvalidType(DXAPIError):
    ''' Raised when an object specified in the request is of invalid type. '''

class RateLimitConditional(DXAPIError):
    ''' Raised when the rate of invalid requests is too high. '''

class InternalError(DXAPIError):
    ''' Raised when the server encountered an internal error. '''

class ServiceUnavailable(DXAPIError):
    ''' Raised when an API service was temporarily unavailable. '''

class DXFileError(DXError):
    '''Exception for :class:`dxpy.bindings.dxfile.DXFile`.'''

class DXIncompleteReadsError(DXError):
    '''Exception for :class:`dxpy.bindings.dxfile.DXFile` when returned read data is shorter than requested'''

class DXPartLengthMismatchError(DXFileError):
    '''Exception raised by :class:`dxpy.bindings.dxfile.DXFile` on part length mismatch.'''

class DXChecksumMismatchError(DXFileError):
    '''Exception raised by :class:`dxpy.bindings.dxfile.DXFile` on checksum mismatch.'''

class DXSearchError(DXError):
    '''Exception for :mod:`dxpy.bindings.search` methods.'''

class DXAppletError(DXError):
    '''Exception for :class:`dxpy.bindings.dxapplet.DXApplet`.'''

class DXJobFailureError(DXError):
    '''Exception produced by :class:`dxpy.bindings.dxjob.DXJob` when a job fails.'''

class ProgramError(DXError):
    '''Deprecated. Use :class:`AppError` instead.'''

class AppError(ProgramError):
    '''
    Base class for fatal exceptions to be raised while using :mod:`dxpy` inside
    DNAnexus execution containers.

    This exception is thrown for user errors, and the error message is
    presented to the user. Throwing this exception will cause the Python
    execution template to write exception information into the file
    *job_error.json* in the current working directory, allowing reporting of
    the error state through the DNAnexus API.
    '''

class AppInternalError(DXError):
    '''
    Base class for fatal exceptions to be raised while using :mod:`dxpy` inside
    DNAnexus execution containers.

    This exception is intended for internal App errors, whose message goes to
    the App developer. Throwing this exception will cause the Python execution
    template to write exception information into the file ``job_error.json`` in
    the current working directory, allowing reporting of the error state
    through the DNAnexus API.
    '''

class DXCLIError(DXError):
    '''
    Exception class for generic errors in the command-line client
    '''

class ContentLengthError(HTTPError):
    '''
    Raised when actual content length received from the server does not
    match the "Content-Length" header
    '''

class HTTPErrorWithContent(HTTPError):
    '''
    Specific variant of HTTPError with response content.

    This class was created to avoid appending content directly to error message
    which makes difficult to format log strings.
    '''

    def __init__(self, value, content):
        super(HTTPError, self).__init__(value)
        self.content = content

class BadJSONInReply(ValueError):
    '''
    Raised when the server returned invalid JSON in the response body. Possible reasons
    for this are the network connection breaking, or overload on the server.
    '''

class InvalidTLSProtocol(DXAPIError):
    '''
    Raised when the connection to the server was reset due to an ssl protocol not supported.
    Only connections with TLS v1.2 will be accepted.
    '''
    def __init__(self):
        pass

    def error_message(self):
        output = "Please refer to our blog post at https://blog.dnanexus.com/2017-09-23-upgrading-tls/ regarding upgrading to TLS 1.2."
        return output

    def __str__(self):
        return self.error_message()

class UrllibInternalError(AttributeError):
    '''
    Exception class for AttributeError from urllib3
    '''


def format_exception(e):
    """Returns a string containing the type and text of the exception.

    """
    from .utils.printing import fill
    return '\n'.join(fill(line) for line in traceback.format_exception_only(type(e), e))


def exit_with_exc_info(code=1, message='', print_tb=False, exception=None):
    '''Exits the program, printing information about the last exception (if
    any) and an optional error message.  Uses *exception* instead if provided.

    :param code: Exit code.
    :type code: integer (valid exit code, 0-255)
    :param message: Message to be printed after the exception information.
    :type message: string
    :param print_tb: If set to True, prints the exception traceback; otherwise, suppresses it.
    :type print_tb: boolean
    :type exception: an exception to use in place of the last exception raised
    '''
    exc_type, exc_value = (exception.__class__, exception) \
                          if exception is not None else sys.exc_info()[:2]

    if exc_type is not None:
        if print_tb:
            traceback.print_exc()
        elif isinstance(exc_value, KeyboardInterrupt):
            sys.stderr.write('^C\n')
        else:
            for line in traceback.format_exception_only(exc_type, exc_value):
                sys.stderr.write(line)

    sys.stderr.write(message)
    if message != '' and not message.endswith('\n'):
        sys.stderr.write('\n')
    sys.exit(code)

network_exceptions = (urllib3.exceptions.ProtocolError,
                      urllib3.exceptions.NewConnectionError,
                      urllib3.exceptions.DecodeError,
                      urllib3.exceptions.ConnectTimeoutError,
                      urllib3.exceptions.ReadTimeoutError,
                      urllib3.connectionpool.HTTPException,
                      urllib3.exceptions.SSLError,
                      ssl.SSLError,
                      HTTPError,
                      socket.error,
                      ConnectionResetError,)

try:
    json_exceptions = (json.decoder.JSONDecodeError,)
except:
    json_exceptions = (ValueError,)

default_expected_exceptions = network_exceptions + json_exceptions + (DXAPIError,
                                                    DXCLIError,
                                                    KeyboardInterrupt)

def err_exit(message='', code=None, expected_exceptions=default_expected_exceptions, arg_parser=None,
             ignore_sigpipe=True, exception=None):
    '''Exits the program, printing information about the last exception (if
    any) and an optional error message.  Uses *exception* instead if provided.

    Uses **expected_exceptions** to set the error code decide whether to
    suppress the error traceback.

    :param message: Message to be printed after the exception information.
    :type message: string
    :param code: Exit code.
    :type code: integer (valid exit code, 0-255)
    :param expected_exceptions: Exceptions for which to exit with error code 3 (expected error condition) and suppress the stack trace (unless the _DX_DEBUG environment variable is set).
    :type expected_exceptions: iterable
    :param arg_parser: argparse.ArgumentParser object used in the program (optional)
    :param ignore_sigpipe: Whether to exit silently with code 3 when IOError with code EPIPE is raised. Default true.
    :type ignore_sigpipe: boolean
    :param exception: an exception to use in place of the last exception raised
    '''
    if arg_parser is not None:
        message = arg_parser.prog + ": " + message

    exc = exception if exception is not None else sys.exc_info()[1]
    if isinstance(exc, SystemExit):
        raise exc
    elif isinstance(exc, expected_exceptions):
        exit_with_exc_info(EXPECTED_ERR_EXIT_STATUS, message, print_tb=dxpy._DEBUG > 0, exception=exception)
    elif ignore_sigpipe and isinstance(exc, IOError) and getattr(exc, 'errno', None) == errno.EPIPE:
        if dxpy._DEBUG > 0:
            print("Broken pipe", file=sys.stderr)
        sys.exit(3)
    else:
        if code is None:
            code = 1
        exit_with_exc_info(code, message, print_tb=True, exception=exception)



# Copyright (C) 2013-2019 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

'''
DXDatabase Handler
**************

This remote database handler is a Python database-like object.
'''

from __future__ import print_function, unicode_literals, division, absolute_import

import os, copy, time
from io import BytesIO
from threading import Lock
from multiprocessing import cpu_count

import dxpy
from . import DXDataObject
from ..utils.resolver import object_exists_in_project
from .. import logger
from ..compat import basestring


DXFILE_HTTP_THREADS = min(cpu_count(), 8)
MIN_BUFFER_SIZE = 1024*1024
DEFAULT_BUFFER_SIZE = 1024*1024*16
if dxpy.JOB_ID:
    # Increase HTTP request buffer size when we are running within the
    # platform.
    DEFAULT_BUFFER_SIZE = 1024*1024*96

MD5_READ_CHUNK_SIZE = 1024*1024*4
FILE_REQUEST_TIMEOUT = 60


def _validate_headers(headers):
    for key, value in headers.items():
        if not isinstance(key, basestring):
            raise ValueError("Expected key %r of headers to be a string" % (key,))
        if not isinstance(value, basestring):
            raise ValueError("Expected value %r of headers (associated with key %r) to be a string"
                             % (value, key))
    return headers


def _readable_part_size(num_bytes):
    "Returns the file size in readable form."
    B = num_bytes
    KB = float(1024)
    MB = float(KB * 1024)
    GB = float(MB * 1024)
    TB = float(GB * 1024)

    if B < KB:
        return '{0} {1}'.format(B, 'bytes' if B != 1 else 'byte')
    elif KB <= B < MB:
        return '{0:.2f} KiB'.format(B/KB)
    elif MB <= B < GB:
        return '{0:.2f} MiB'.format(B/MB)
    elif GB <= B < TB:
        return '{0:.2f} GiB'.format(B/GB)
    elif TB <= B:
        return '{0:.2f} TiB'.format(B/TB)

def do_debug(msg):
    logger.debug(msg)

class DXDatabase(DXDataObject):
    '''Remote database object handler.

    :param dxid: Object ID
    :type dxid: string
    :param project: Project ID
    :type project: string

    '''

    _class = "database"

    _describe = staticmethod(dxpy.api.file_describe)

    _http_threadpool_size = DXFILE_HTTP_THREADS
    _http_threadpool = dxpy.utils.get_futures_threadpool(max_workers=_http_threadpool_size)

    NO_PROJECT_HINT = 'NO_PROJECT_HINT'

    def __init__(self, dxid=None, project=None, read_buffer_size=DEFAULT_BUFFER_SIZE,
                 expected_file_size=None, file_is_mmapd=False):
        """
        :param dxid: Object ID
        :type dxid: string
        :param project: Project ID
        :type project: string
        :param read_buffer_size: size of read buffer in bytes
        :type read_buffer_size: int
        :param expected_file_size: size of data that will be written, if
            known
        :type expected_file_size: int
        :param file_is_mmapd: True if input file is mmap'd (if so, the
            write buffer size will be constrained to be a multiple of
            the allocation granularity)
        :type file_is_mmapd: bool
        """

        DXDataObject.__init__(self, dxid=dxid, project=project)

        self._read_buf = BytesIO()
        self._read_bufsize = read_buffer_size
        self._expected_file_size = expected_file_size
        self._file_is_mmapd = file_is_mmapd

        # These are cached once for all download threads. This saves calls to the apiserver.
        self._download_url, self._download_url_headers, self._download_url_expires = None, None, None

        # This lock protects accesses to the above three variables, ensuring that they would
        # be checked and changed atomically. This protects against thread race conditions.
        self._url_download_mutex = Lock()

        self._request_iterator, self._response_iterator = None, None
        self._http_threadpool_futures = set()


    def get_download_url(self, filename=None, src_filename=None, project=None, **kwargs):
        """
        :param filename: desired filename of the local downloaded file
        :type filename: str
        :param project: ID of a project containing the file (the download URL
            will be associated with this project, and this may affect which
            billing account is billed for this download).
            If no project is specified, an attempt will be made to verify if the database is
            in the project from the DXDatabase handler (as specified by the user or
            the current project stored in dxpy.WORKSPACE_ID). Otherwise, no hint is supplied.
            This fall back behavior does not happen inside a job environment.
        :type project: str
        :returns: download URL and dict containing HTTP headers to be supplied
            with the request
        :rtype: tuple (str, dict)
        :raises: :exc:`~dxpy.exceptions.ResourceNotFound` if a project context was
            given and the database was not found in that project context.
        :raises: :exc:`~dxpy.exceptions.ResourceNotFound` if no project context was
            given and the database was not found in any projects.

        Obtains a URL that can be used to directly download files associated
        with the specified database.

        """

        args = {}

        if src_filename is not None:
            args["filename"] = src_filename

        # If project=None, we fall back to the project attached to this handler
        # (if any). If this is supplied, it's treated as a hint: if it's a
        # project in which this file exists, it's passed on to the
        # apiserver. Otherwise, NO hint is supplied. In principle supplying a
        # project in the handler that doesn't contain this file ought to be an
        # error, but it's this way for backwards compatibility. We don't know
        # who might be doing downloads and creating handlers without being
        # careful that the project encoded in the handler contains the file
        # being downloaded. They may now rely on such behavior.
        if project is None and 'DX_JOB_ID' not in os.environ:
            project_from_handler = self.get_proj_id()
            if project_from_handler and object_exists_in_project(self.get_id(), project_from_handler):
                project = project_from_handler

        if project is not None and project is not DXDatabase.NO_PROJECT_HINT:
            args["projectContext"] = project

        # Test hook to write 'project' argument passed to API call to a
        # local file
        if '_DX_DUMP_BILLED_PROJECT' in os.environ:
            with open(os.environ['_DX_DUMP_BILLED_PROJECT'], "w") as fd:
                if project is not None and project != DXDatabase.NO_PROJECT_HINT:
                    fd.write(project)

        with self._url_download_mutex:

            if self._download_url is None or self._download_url_expires < time.time():
                # The idea here is to cache a download URL for the entire file, that will
                # be good for a few minutes. This avoids each thread having to ask the
                # server for a URL, increasing server load.
                #
                # To avoid thread race conditions, this check/update procedure is protected
                # with a lock.

                # logging.debug("Download URL unset or expired, requesting a new one")
                if "timeout" not in kwargs:
                    kwargs["timeout"] = FILE_REQUEST_TIMEOUT
                do_debug("dxdatabase get_download_url - download_file args = {}".format(args))
                resp = dxpy.api.database_download_file(self._dxid, args, **kwargs)
                do_debug("dxdatabase get_download_url - download_file resp = {}".format(resp));
                self._download_url = resp["url"]
                self._download_url_headers = _validate_headers(resp.get("headers", {}))
                self._download_url_expires = 32503680000  # doesn't expire (year 3000)

            # Make a copy, ensuring each thread has its own mutable
            # version of the headers.  Note: python strings are
            # immutable, so we can safely give a reference to the
            # download url.
            retval_download_url = self._download_url
            retval_download_url_headers = copy.copy(self._download_url_headers)

        return retval_download_url, retval_download_url_headers



#!/usr/bin/env python3
# DX_APP_WIZARD_NAME DX_APP_WIZARD_VERSION
# Generated by dx-app-wizard.
#
# Basic execution pattern: Your app will run on a single machine from
# beginning to end.
#
# See https://documentation.dnanexus.com/developer for documentation and
# tutorials on how to modify this file.
#
# DNAnexus Python Bindings (dxpy) documentation:
#   http://autodoc.dnanexus.com/bindings/python/current/

import os
import dxpy

@dxpy.entry_point('main')
def main(DX_APP_WIZARD_INPUT_SIGNATURE):
DX_APP_WIZARD_INITIALIZE_INPUTDX_APP_WIZARD_DOWNLOAD_ANY_FILES
    # Fill in your application code here.
DX_APP_WIZARD_UPLOAD_ANY_FILES
    # The following line fills in some basic dummy output and assumes
    # that you have created variables to represent your output with
    # the same name as your output fields.

    output = {}
DX_APP_WIZARD_OUTPUT
    return output

dxpy.run()



# Copyright (C) 2013-2016 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""
Utilities for client-side usage of the streaming log API
(https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-job-xxxx-getlog).
"""

from __future__ import print_function, unicode_literals, division, absolute_import

import json
import logging
import os
import signal
import ssl
import sys
import textwrap
import time

from threading import Thread
from websocket import WebSocketApp

import dxpy
from .describe import get_find_executions_string
from ..exceptions import err_exit

logger = logging.getLogger('websocket')
logger.setLevel(logging.WARN)


class DXJobLogStreamingException(Exception):
    pass


class DXJobLogStreamClient:
    def __init__(
        self, job_id, job_try=None, input_params=None, msg_output_format="{job} {level} {msg}",
        msg_callback=None, print_job_info=True, exit_on_failed=True
    ):
        """Initialize job log client.

        :param job_id: dxid for a job (hash ID 'job-xxxx')
        :type job_id: str
        :param job_try: try for given job. If None, it will use the latest try.
        :type job_id: int or None
        :param input_params: blob with connection parameters, should have keys
        ``numRecentMessages`` (int) (wich may not be more than 1024 * 256, otherwise no logs will be returned),
        ``recurseJobs`` (bool) - if True, attempts to traverse subtree
        ``tail`` (bool) - if True, keep watching job. Should also be set to True to get all logs
        from a completed job.
        :type input_params: dict
        :param msg_output_format: how messages should be printed to console. Ignored if
        ``msg_callback`` is specified
        :type msg_output_form: str
        :param print_job_info: if True, prints metadata about job
        :type print_job_info: bool
        :param msg_callback: single argument function that accepts a JSON blob with message
        details. Example:
        ``{"timestamp": 1575465039481, "source": "APP", "level": "STDOUT", "job": "job-123",
           "line":24, "msg": "success WfFragment"}``
        where ``timestamp`` is Unix epoch time in milliseconds and ``line`` is a sequence number.
        :type msg_callback: callable
        :param exit_on_failed: if True, will raise SystemExit with code of 3 if encountering a
        failed job (this is the default behavior)
        :type exit_on_failed: bool
        """
        # TODO: add unit tests; note it is a public class

        self.job_id = job_id
        self.job_try = job_try
        self.job_has_try = job_try is not None
        self.input_params = input_params
        self.msg_output_format = msg_output_format
        self.msg_callback = msg_callback
        self.print_job_info = print_job_info
        self.seen_jobs = {}
        self.error = False
        self.exception = None
        self.closed_code = None
        self.closed_reason = None
        self.exit_on_failed = exit_on_failed
        self.url = "{protocol}://{host}:{port}/{job_id}/getLog/websocket".format(
            protocol='wss' if dxpy.APISERVER_PROTOCOL == 'https' else 'ws',
            host=dxpy.APISERVER_HOST,
            port=dxpy.WATCH_PORT if dxpy.WATCH_PORT is not None else dxpy.APISERVER_PORT,
            job_id=job_id
        )
        self._app = None

    def connect(self):
        while True:
            self.error = False
            self.exception = None
            self.closed_code = None
            self.closed_reason = None

            try:
                self._app = WebSocketApp(
                    self.url,
                    on_open=lambda app: self.opened(),
                    on_close=lambda app, close_status_code, close_msg: self.closed(close_status_code, close_msg),
                    on_error=lambda app, exception: self.errored(exception),
                    on_message=lambda app, message: self.received_message(message)
                )
                self._app.run_forever(sslopt={"cert_reqs": ssl.CERT_NONE})
            except:
                if not self.server_restarted():
                    raise
            finally:
                self._app = None

            if self.server_restarted():
                # Instead of trying to reconnect in a retry loop with backoff, run an
                # API call that will do the same and block while it retries.
                logger.warn("Server restart, reconnecting...")
                time.sleep(1)
                self._describe_job(self.job_id)
            else:
                break

    def server_restarted(self):
        return (
            self.closed_code == 1001 and
            self.closed_reason == "Server restart, please reconnect later"
        )

    def opened(self):
        args = {
            "access_token": dxpy.SECURITY_CONTEXT['auth_token'],
            "token_type": dxpy.SECURITY_CONTEXT['auth_token_type']
        }
        if self.input_params:
            args.update(self.input_params)
        self._app.send(json.dumps(args))

    def errored(self, exception=None):
        self.error = True
        self.exception = exception

    def closed(self, code=None, reason=None):
        if code:
            self.closed_code = code
            self.closed_reason = reason
        elif not self.error:
            self.closed_code = 1000
            self.closed_reason = "Normal"
        elif self.exception and type(self.exception) in {KeyboardInterrupt, SystemExit}:
            self.closed_code = 1000
            self.closed_reason = "Connection terminated by client"
        else:
            self.closed_code = 1006
            self.closed_reason = str(self.exception) if self.exception else "Abnormal"

        if self.closed_code != 1000:
            try:
                error = json.loads(self.closed_reason)
                raise DXJobLogStreamingException(
                    "Error while streaming job logs: {type}: {message}\n".format(
                        **error
                    )
                )
            except (KeyError, ValueError):
                raise DXJobLogStreamingException(
                    "Error while streaming job logs: {code}: {reason}\n".format(
                        code=self.closed_code, reason=self.closed_reason
                    )
                )
        elif self.print_job_info:
            if self.job_id not in self.seen_jobs:
                self.seen_jobs[self.job_id] = {}
            for job_id in self.seen_jobs.keys():
                self.seen_jobs[job_id] = self._describe_job(job_id)
                print(
                    get_find_executions_string(
                        self.seen_jobs[job_id],
                        has_children=False,
                        show_outputs=True,
                        show_try=self.job_has_try
                    )
                )
        else:
            self.seen_jobs[self.job_id] = self._describe_job(self.job_id)

        if (self.exit_on_failed
                and self.seen_jobs[self.job_id].get('state') in {'failed', 'terminated'}):
            err_exit(code=3)

    def received_message(self, message):
        message_dict = json.loads(message)

        if (
            self.print_job_info and
            'job' in message_dict and
            message_dict['job'] not in self.seen_jobs
        ):
            self.seen_jobs[message_dict['job']] = self._describe_job(message_dict['job'])
            print(
                get_find_executions_string(
                    self.seen_jobs[message_dict['job']],
                    has_children=False,
                    show_outputs=False,
                    show_try=self.job_has_try
                )
            )

        if (
            message_dict.get('source') == 'SYSTEM' and
            message_dict.get('msg') == 'END_LOG'
        ):
            self._app.keep_running = False
        elif self.msg_callback:
            self.msg_callback(message_dict)
        else:
            print(self.msg_output_format.format(**message_dict))

    def _describe_job(self, job_id):
        return dxpy.api.job_describe(job_id, {'try': self.job_try} if self.job_has_try else {})


class CursesDXJobLogStreamClient(DXJobLogStreamClient):

    def closed(self, *args, **kwargs):
        super(CursesDXJobLogStreamClient, self).closed(args, kwargs)
        # Overcome inability to stop Python process from a thread by sending SIGINT
        os.kill(os.getpid(), signal.SIGINT)


def metrics_top(args, input_params, enrich_msg):
    try:
        import curses
    except:
        err_exit("--metrics top is not supported on your platform due to missing curses library")

    class ScreenManager:

        def __init__(self, args):
            self.stdscr = None
            self.args = args
            self.log_client = CursesDXJobLogStreamClient(args.jobid, input_params=input_params, msg_callback=self.msg_callback,
                                                         msg_output_format=None, print_job_info=False, exit_on_failed=False)
            self.curr_screen = 'logs'
            self.log = []
            self.metrics = ['Waiting for job logs...']
            self.scr_dim_y = 0
            self.scr_y_offset = 0
            self.scr_y_max_offset = 0
            self.scr_dim_x = 0
            self.scr_x_offset = 0
            self.scr_x_max_offset = 0
            self.curr_row = 0
            self.curr_row_total_chars = 0
            self.curr_col = 0

        def main(self, stdscr):
            self.stdscr = stdscr

            curses.use_default_colors()
            curses.init_pair(1, curses.COLOR_BLUE, -1)
            curses.init_pair(2, curses.COLOR_RED, -1)
            curses.init_pair(3, curses.COLOR_YELLOW, -1)
            curses.init_pair(4, curses.COLOR_GREEN, -1)

            t = Thread(target=self.log_client.connect)
            t.daemon = True
            t.start()

            self.refresh()
            try:
                while True:
                    ch = stdscr.getch()
                    if ch == curses.KEY_RESIZE: self.refresh()
                    elif self.curr_screen == 'logs':
                        if ch == curses.KEY_RIGHT: self.refresh(scr_x_offset_diff=1)
                        elif ch == curses.KEY_LEFT: self.refresh(scr_x_offset_diff=-1)
                        elif ch == curses.KEY_SRIGHT: self.refresh(scr_x_offset_diff=20)
                        elif ch == curses.KEY_SLEFT: self.refresh(scr_x_offset_diff=-20)
                        elif ch == curses.KEY_HOME: self.refresh(scr_x_offset_diff=-self.scr_x_offset)
                        elif ch == curses.KEY_END: self.refresh(scr_x_offset_diff=self.scr_x_max_offset)
                        elif ch == curses.KEY_UP: self.refresh(scr_y_offset_diff=1)
                        elif ch == curses.KEY_DOWN: self.refresh(scr_y_offset_diff=-1)
                        elif ch == curses.KEY_PPAGE: self.refresh(scr_y_offset_diff=10)
                        elif ch == curses.KEY_NPAGE: self.refresh(scr_y_offset_diff=-10)
                        elif ch == ord('?') or ch == ord('?'): self.refresh(target_screen='help')
                        elif ch == ord('q') or ch == ord('Q'): sys.exit(0)
                    elif self.curr_screen == 'help':
                        if ch >= 0: self.refresh(target_screen='logs')
            # Capture SIGINT and exit normally
            except KeyboardInterrupt:
                sys.exit(0)

        def msg_callback(self, message):
            if len(self.log) == 0:
                self.metrics[0] = ''

            enrich_msg(self.log_client, message)
            if message['level'] == 'METRICS':
                self.metrics[0] = '[%s] %s' % (message['timestamp'], message['msg'])
            else:
                self.log.append(message)
                if self.scr_y_offset > 0:
                    self.scr_y_offset += 1

            self.refresh()

        def refresh(self, target_screen=None, scr_y_offset_diff=None, scr_x_offset_diff=None):
            self.stdscr.erase()
            self.scr_dim_y, self.scr_dim_x = self.stdscr.getmaxyx()

            self.scr_y_max_offset = max(len(self.log) - self.scr_dim_y + 3, 0)
            self.update_screen_offsets(scr_y_offset_diff, scr_x_offset_diff)

            self.curr_row = 0

            if target_screen is not None:
                self.curr_screen = target_screen

            if self.curr_screen == 'help':
                self.draw_help()
            else:
                self.draw_logs()

            self.stdscr.refresh()

        def draw_logs(self):
            nlines = min(self.scr_dim_y - 3, len(self.log))
            self.stdscr.addnstr(self.curr_row, 0, self.metrics[-1], self.scr_dim_x)
            self.curr_row += 2

            for i in range(nlines):
                message = self.log[len(self.log) - nlines + i - self.scr_y_offset]
                self.curr_col = 0
                self.curr_row_total_chars = 0

                if args.format:
                    self.print_field(args.format.format(**message), 0)
                else:
                    if self.args.timestamps:
                        self.print_field(message['timestamp'], 0)
                    self.print_field(message['job_name'], 1)
                    if self.args.job_ids:
                        self.print_field('(%s)' % message['job'], 1)
                    self.print_field(message.get('level', ''), message['level_color_curses'])
                    self.print_field(message['msg'], 0)

                self.scr_x_max_offset = max(self.scr_x_max_offset, self.curr_row_total_chars - 1)
                self.curr_row += 1

        def draw_help(self):
            text = '''Metrics top mode help
_
This mode shows the latest METRICS message at the top of the screen and updates it for running jobs instead of showing every METRICS message interspersed with the currently-displayed job log messages. For completed jobs, this mode does not show any metrics.
_
Controls:
  Up/Down               scroll up/down by one line
  PgUp/PgDn             scroll up/down by 10 lines
  Left/Right            scroll left/right by one character
  Shift + Left/Right    scroll left/right by 20 characters
  Home/End              scroll to the beginning/end of the line
  ?                     display this help
  q                     quit
_
Press any key to return.
'''
            lines = []
            for line in text.splitlines():
                if line == '_':
                    lines.append('')
                    continue
                lines += textwrap.wrap(line, self.scr_dim_x - 1)

            for row in range(min(len(lines), self.scr_dim_y)):
                self.stdscr.addnstr(row, 0, lines[row], self.scr_dim_x)

        def print_field(self, text, color):
            if self.curr_col < self.scr_dim_x:
                if self.curr_row_total_chars >= self.scr_x_offset:
                    self.stdscr.addnstr(self.curr_row, self.curr_col, text, self.scr_dim_x - self.curr_col, curses.color_pair(color))
                    self.curr_col += len(text) + 1
                elif self.curr_row_total_chars + len(text) + 1 > self.scr_x_offset:
                    self.stdscr.addnstr(self.curr_row, self.curr_col, text[self.scr_x_offset - self.curr_row_total_chars:], self.scr_dim_x - self.curr_col, curses.color_pair(color))
                    self.curr_col += len(text[self.scr_x_offset - self.curr_row_total_chars:]) + 1

            self.curr_row_total_chars += len(text) + 1

        def update_screen_offsets(self, diff_y, diff_x):
            if not diff_y:
                diff_y = 0
            if not diff_x:
                diff_x = 0
            self.scr_y_offset = min(self.scr_y_offset + diff_y, self.scr_y_max_offset) if diff_y > 0 else max(self.scr_y_offset + diff_y, 0)
            self.scr_x_offset = min(self.scr_x_offset + diff_x, self.scr_x_max_offset) if diff_x > 0 else max(self.scr_x_offset + diff_x, 0)

    manager = ScreenManager(args)
    curses.wrapper(manager.main)



from __future__ import print_function


class JSONFiltersValidator(object):
    """
    A specialized class that parses user input JSON according to a schema to prepare vizserver-compliant compound filters.

    See assay_filtering_conditions.py for current schemas, which include optimized schema EXTRACT_ASSAY_EXPRESSION_FILTERING_CONDITIONS_1_1
    and non-optimized EXTRACT_ASSAY_EXPRESSION_FILTERING_CONDITIONS_1_1_non_optimized to handle location filters.

    Filters must be defined in schema["filtering_conditions"]

    There are currently three ways to define filtering_conditions when version is 1.1:
    1. Basic use-case: no "properties" are defined. Only "condition", "table_column" and optionally "max_item_limit" are defined.
                       In this case, there are no sub-keys for the current key in input_json.
                       (see 'sample_id' in 'assay_filtering_conditions.EXTRACT_ASSAY_EXPRESSION_FILTERING_CONDITIONS_1_1')
    2. Properties defined as dict of dicts (see 'annotation' and 'expression' in EXTRACT_ASSAY_EXPRESSION_FILTERING_CONDITIONS_1_1)
        2.1. If filters_combination_operator is not defined, the assumption is that there is only one sub-key in input_json

    3. Complex use-case: Properties defined as list of dicts (more advanced, special use-case with complex conditional logics that needs translation)
        Filtering conditions defined this way indicate that input_json may contain more than one item for the current key, and elements are stored in a list.
        In this case, the schema must define "items_combination_operator" and "filters_combination_operator".
        items_combination_operator: how to combine filters for list items in input_json
        filters_combination_operator: how to combine filters within each item
        (see 'location' in EXTRACT_ASSAY_EXPRESSION_FILTERING_CONDITIONS_1_1_non_optimized)

        Within 'properties' if "key" is defined, then a generic one-key filter is built.
        If "keys" is defined, then more complex use-cases are handled via special conditions that are defined in condition_function_mapping.

    """

    def __init__(self, input_json, schema, error_handler=print):
        self.input_json = input_json
        self.schema = schema
        self.error_handler = error_handler
        self.condition_function_mapping = {
            "genobin_partial_overlap": self.build_partial_overlap_genobin_filters,
        }
        self.SUPPORTED_VIZSERVER_CONDITIONS = [
            "contains",
            "exists",
            "not-exists",
            "any",
            "not-any",
            "all",
            "not-empty",
            "in",
            "not-in",
            "is",
            "is-not",
            "greater-than",
            "less-than",
            "greater-than-eq",
            "less-than-eq",
            "between",
            "between-ex",
            "between-left-inc",
            "between-right-inc",
            "not-between",
            "not-between-ex",
            "not-between-left-inc",
            "not-between-right-inc",
            "compare-before",
            "compare-after",
            "compare-within",
        ]

    def parse(self):
        self.is_valid_json(self.schema)
        if self.get_schema_version(self.schema).startswith("1."):
            return self.parse_v1_1()
        else:
            raise NotImplementedError

    def parse_v1_1(self):
        """
        Parse input_json according to schema version 1.1 (optimized), and build vizserver compound filters. In this version,
        compound filters are built in one hash.
        """
        return self.merge_duplicate_filters(self.parse_v1())

    def parse_v1(self):
        """
        Parse input_json according to schema version 1.0 (non-optimized), and build vizserver compound filters.
        """

        try:
            # Get the general structure of vizserver-compliant compound filter dict
            vizserver_compound_filters = self.get_vizserver_basic_filter_structure()

            # Get 'filtering_conditions' from the schema
            all_filters = self.collect_filtering_conditions(self.schema)

            # Go through the input_json (iterate through keys and values in user input JSON)
            for filter_key, filter_values in self.input_json.items():
                # Example: if schema is EXTRACT_ASSAY_EXPRESSION_FILTERING_CONDITIONS_1_1_non_optimized
                # Then input JSON would contain location and maybe others such as annotation
                # In this case:
                # filter_key -> location
                # filter_values -> list of dicts (where each dict contains "chromosome", "starting_position", "ending_position")
                if filter_key not in all_filters:
                    self.error_handler(
                        "No filtering condition was defined for {} in the schema.".format(
                            filter_key
                        )
                    )

                # Get the filtering conditions for the current "key" in input_json
                current_filters = all_filters[filter_key]

                # Get the properties if any (this might be None,
                # if so we will just execute the basic use case defined in the docstring of the class)
                current_properties = current_filters.get("properties")

                # Validate max number of allowed items
                # if max_item_limit is defined at the top level within this key
                # It will be later validated for each property as well

                self.validate_max_item_limit(current_filters, filter_values, filter_key)

                # There are several ways filtering_conditions can be defined
                # 1. Basic use-case: no properties, just condition (see 'sample_id' in 'EXTRACT_ASSAY_EXPRESSION_FILTERING_CONDITIONS_1_1')
                # 2. Properties defined as dict of dicts (see 'annotation' and 'expression')
                # 3. Properties defined as list of dicts (more advanced, special use-case with complex conditional logics that needs translation)
                # For more information see the docstring of the class
                # The following if conditions will go through each of the aforementioned scenarios

                if isinstance(current_properties, list) and isinstance(
                    filter_values, list
                ):
                    filters = self.parse_list_v1(
                        current_filters,
                        filter_values,
                        current_properties,
                    )

                    if filters is not None:
                        vizserver_compound_filters["compound"].append(filters)
                        filters = None

                if isinstance(current_properties, dict):
                    for k, v in current_properties.items():
                        if k in filter_values:
                            self.validate_max_item_limit(v, filter_values[k], k)

                    filters = self.parse_dict_v1(
                        current_filters, filter_values, current_properties
                    )
                    if filters is not None:
                        vizserver_compound_filters["compound"].append(filters)
                        filters = None

                if current_properties is None:
                    # no properties, so just apply conditions
                    # In other words .get("properties") returns None
                    # Therefore we are dealing with a basic use-case scenario
                    # (See 'sample_id' in 'EXTRACT_ASSAY_EXPRESSION_FILTERING_CONDITIONS_1_1' for an example)
                    filters = self.build_one_key_generic_filter(
                        current_filters.get("table_column"),
                        current_filters.get("condition"),
                        filter_values,
                    )

                    if filters is not None:
                        vizserver_compound_filters["compound"].append(filters)
                        filters = None

            return vizserver_compound_filters

        except Exception as e:
            self.error_handler(str(e))

    def parse_list_v1(self, current_filters, filter_values, current_properties):
        if current_properties is None:
            self.error_handler("Expected properties to be defined within schema.")

        # There are two important aspects:
        # Filters (if more than one) must be compounded for each item with filters_combination_operator logic
        # All items must be compounded together as one large compounded filter with the logic defined in items_combination_operator

        full_filter_for_all_items = {
            "logic": current_filters.get("items_combination_operator"),
            "compound": [
                # {base_filter_for_each_item_1},
                # {base_filter_for_each_item_2},
                # {etc.}
                # as many dicts as there are "items" in input_json[filter_key]
            ],
        }

        ### current_properties is a list of dicts and each dict is a filter
        ### therefore, we need to iterate over each dict and build a filter for each one
        ### However, also remember than inside filter_values from input_json we have a list of dicts
        ### therefore there might be more than one element within that list
        ### so we need to build the filters for each element in the list as well
        ### and then append them to the compound list of dicts

        # Build filters for each item in the list (input_json[filter_values])
        for current_list_item in filter_values:
            # Consider keeping track of input_json keys and properties evaluated and used so far

            current_compound_filter = {
                "logic": current_filters.get("filters_combination_operator"),
                "compound": [
                    # {},
                    # {},
                    # here there will be as many dicts inside as there are qualifying [key] conditions
                ],
            }

            ## properties['key'] -> simple case
            ## properties['keys'] -> reserved for complex, special cases
            ## specifically for those cases where SUPPORTED_VIZSERVER_CONDITIONS is not sufficient

            for item in current_properties:
                if item.get("key"):
                    # Consider a separate check for list type
                    # Currently not necessary
                    temp_filter = self.build_one_key_generic_filter(
                        item["table_column"],
                        item["condition"],
                        current_list_item[item.get("key")],
                    )

                    current_compound_filter["compound"].append(temp_filter)

                if item.get("keys"):
                    if len(item.get("keys")) == 2:
                        if (
                            item.get("condition")
                            not in self.SUPPORTED_VIZSERVER_CONDITIONS
                        ):
                            special_filtering_function = (
                                self.condition_function_mapping[item.get("condition")]
                            )
                            temp_filter = special_filtering_function(
                                item, current_list_item
                            )
                            current_compound_filter["compound"].append(temp_filter)

            # Consider checking if current_compound_filter contains any new elements before appending
            full_filter_for_all_items["compound"].append(current_compound_filter)

        return full_filter_for_all_items

    def parse_dict_v1(self, current_filters, filter_values, current_properties):
        if current_properties is None:
            self.error_handler("Expected properties to be defined within schema.")

        # if no filters_combination_operator is specified, then single filter assumed

        filtering_logic = current_filters.get("filters_combination_operator")

        # if filtering_logic isn't defined at this level then there must be only one 'key' to filter on
        if filtering_logic:
            if len(filter_values) == 1:
                temp_key = next(iter(filter_values.keys()))
                matched_filter = current_properties[temp_key]
                temp_filter = self.build_one_key_generic_filter(
                    matched_filter.get("table_column"),
                    matched_filter.get("condition"),
                    filter_values.get(temp_key),
                )
                return temp_filter

            if len(filter_values) == 2:
                temp_keys = list(filter_values.keys())
                first_key = current_properties[temp_keys[0]]
                second_key = current_properties[temp_keys[1]]

                # check if there are two filtering conditions that need to be applied on the same table_column
                # check if both of those are defined in input_json
                # an example of such a case is 'expression' in EXTRACT_ASSAY_EXPRESSION_FILTERING_CONDITIONS_1_1
                # where we might have a `max_value` and a `min_value`
                # however providing both is not mandatory

                if first_key.get("table_column") == second_key.get("table_column"):
                    if filtering_logic == "and":
                        # where possible we will use "between" operator
                        # instead of defining two separate conditions with greater-than AND less-than
                        temp_condition = self.convert_to_between_operator(
                            [
                                first_key.get("condition"),
                                second_key.get("condition"),
                            ]
                        )
                        if temp_condition in ["between", "between-ex"]:
                            temp_values = [
                                filter_values.get(temp_keys[0]),
                                filter_values.get(temp_keys[1]),
                            ]
                            temp_values.sort()
                            temp_filter = self.build_one_key_generic_filter(
                                first_key.get("table_column"),
                                temp_condition,
                                temp_values,
                            )
                            return temp_filter

                        else:
                            # A more complex conversion scenario
                            # that may not involve greater-than/less-than operators
                            # This is currently neither implemented nor needed
                            raise NotImplementedError

                    if filtering_logic == "or":
                        # A special edge case that should probably never be used
                        # It is also not supported by vizserver
                        # In other words, 'or' logic is not supported for
                        # values for the same key in the filter.
                        # If "or" logic needs to be applied to two filters
                        # both with the same 'key', consider constructing two separate
                        # filters and then compounding them with "or" logic
                        raise NotImplementedError

                else:
                    # This is currently not needed anywhere
                    raise NotImplementedError

        else:
            # There's no filtering logic, in other words, filters_combination_operator is not defined at this level
            # (see 'annotation' in 'EXTRACT_ASSAY_EXPRESSION_FILTERING_CONDITIONS_1_1' for an example)
            if len(current_properties) > 1:
                if len(filter_values) > 1:
                    # if there are also more than 1 in input_json
                    self.error_handler(
                        "More than one filter found at this level, but no filters_combination_operator was specified."
                    )
                else:
                    temp_key = next(iter(filter_values.keys()))
                    matched_filter = current_properties[temp_key]
                    temp_filter = self.build_one_key_generic_filter(
                        matched_filter.get("table_column"),
                        matched_filter.get("condition"),
                        filter_values.get(temp_key),
                    )
                return temp_filter

    def get_vizserver_basic_filter_structure(self):
        return {
            "logic": self.get_toplevel_filtering_logic(),
            "compound": [],
        }

    def collect_input_filters(self):
        return self.input_json.keys()

    def get_toplevel_filtering_logic(self):
        return self.schema.get("filters_combination_operator")

    def collect_output_fields_mappings(self, schema):
        return schema.get("output_fields_mapping")

    def collect_filtering_conditions(self, schema):
        return schema.get("filtering_conditions")

    def validate_max_item_limit(self, current, input_json_values, field_name):
        max_item_limit = current.get("max_item_limit")
        if not max_item_limit:
            return True

        if len(input_json_values) > max_item_limit:
            self.error_handler(
                "Too many items given in field {}, maximum is {}.".format(
                    field_name, max_item_limit
                )
            )

    def is_valid_json(self, schema):
        if not isinstance(schema, dict) or not schema:
            self.error_handler("Schema must be a non-empty dict.")

    def get_schema_version(self, schema):
        return schema.get("version")

    def get_number_of_filters(self):
        return len(self.input_json)

    def convert_to_between_operator(self, operators_list):
        if len(operators_list) != 2:
            self.error_handler("Expected exactly two operators")

        if set(operators_list) == {"greater-than-eq", "less-than-eq"}:
            return "between"

        if set(operators_list) == {"greater-than", "less-than"}:
            return "between-ex"

    def build_one_key_generic_filter(
        self, table_column_mapping, condition, value, return_complete_filter=True
    ):
        """
        {
                "filters": {"expr_annotation$chr": [{"condition": "is", "values": 5}]},
        }
        """
        base_filter = {
            table_column_mapping: [{"condition": condition, "values": value}]
        }

        if return_complete_filter:
            return {"filters": base_filter}
        else:
            return base_filter

    def build_partial_overlap_genobin_filters(
        self, filtering_condition, input_json_item
    ):
        """

        Returns a nested compound filter in the following format:

        filter = {
            "logic": "or",
            "compound": [
                {
                    "filters": {
                        db_table_column_start: [
                            {
                                "condition": "between",
                                "values": [input_start_value, input_end_value],
                            }
                        ],
                        db_table_column_end: [
                            {
                                "condition": "between",
                                "values": [input_start_value, input_end_value],
                            }
                        ],
                    },
                    "logic": "or",
                },
                {
                    "filters": {
                        db_table_column_start: [
                            {"condition": "less-than", "values": input_start_value}
                        ],
                        db_table_column_end: [
                            {"condition": "greater-than", "values": input_end_value}
                        ],
                    },
                    "logic": "and",
                },
            ],
        }

        """
        if (
            "starting_position" not in filtering_condition["keys"]
            or "ending_position" not in filtering_condition["keys"]
            or "starting_position" not in input_json_item
            or "ending_position" not in input_json_item
        ):
            self.error_handler(
                "Error in location filtering. starting_position and ending_position must be defined in filtering conditions"
            )

        db_table_column_start = filtering_condition["table_column"]["starting_position"]
        db_table_column_end = filtering_condition["table_column"]["ending_position"]

        input_start_value = int(input_json_item["starting_position"])
        input_end_value = int(input_json_item["ending_position"])

        start_filter = self.build_one_key_generic_filter(
            db_table_column_start,
            "between",
            [input_start_value, input_end_value],
            return_complete_filter=False,
        )
        end_filter = self.build_one_key_generic_filter(
            db_table_column_end,
            "between",
            [input_start_value, input_end_value],
            return_complete_filter=False,
        )
        compound_start_filter = self.build_one_key_generic_filter(
            db_table_column_start,
            "less-than-eq",
            input_start_value,
            return_complete_filter=False,
        )

        compound_end_filter = self.build_one_key_generic_filter(
            db_table_column_end,
            "greater-than-eq",
            input_end_value,
            return_complete_filter=False,
        )

        filter_structure = {
            "logic": "or",
            "compound": [
                {
                    "filters": {},
                    "logic": "or",
                },
                {
                    "filters": {},
                    "logic": "and",
                },
            ],
        }

        filter_structure["compound"][0]["filters"].update(start_filter)
        filter_structure["compound"][0]["filters"].update(end_filter)
        filter_structure["compound"][1]["filters"].update(compound_start_filter)
        filter_structure["compound"][1]["filters"].update(compound_end_filter)

        # In Python 3, this can be simply done via:

        # return {
        #     "logic": "or",
        #     "compound": [
        #         {
        #             "filters": {
        #                 **start_filter,
        #                 **end_filter,
        #             },
        #             "logic": "or",
        #         },
        #         {
        #             "filters": {
        #                 **compound_start_filter,
        #                 **compound_end_filter,
        #             },
        #             "logic": "and",
        #         },
        #     ],
        # }

        return filter_structure

    def merge_duplicate_filters(self, compound_block):
        """
        Recursively walk the compound tree and merge duplicate 'filters' blocks
        at the same compound level.

        Example usage:
        # For
        compound_block = 
                    {
                "logic": "and",
                "compound": [
                    {
                        "filters": {
                            "expression_read_optimized$value": [
                                {"condition": "greater-than-eq", "values": 2000}
                            ]
                        }
                    },
                    {
                        "filters": {
                            "expression_read_optimized$gene_name": [
                                {"condition": "in", "values": ["FAM229B", "TRAIP", "PLEC"]}
                            ]
                        }
                    },
                ],
            }
        merged_block = merge_duplicate_filters(compound_block)

        # merged_block will be:
        {
            "logic": "and",
            "compound": [
                {
                    "filters": {
                        "expression_read_optimized$value": [
                            {"condition": "greater-than-eq", "values": 2000}
                        ],
                        "expression_read_optimized$gene_name": [
                            {"condition": "in", "values": ["FAM229B", "TRAIP", "PLEC"]}
                        ],
                    }
                }
            ],
        }
        """
        if "compound" in compound_block:
            new_compound = []
            merged_filters = {}

            for item in compound_block["compound"]:
                self.merge_duplicate_filters(item)

                if "filters" in item and len(item) == 1:
                    # Merge filters into the accumulating dict
                    for key, conditions in item["filters"].items():
                        merged_filters[key] = []
                        merged_filters[key].extend(conditions)
                else:
                    if merged_filters:
                        new_compound.append({"filters": merged_filters})
                        merged_filters = {}
                    new_compound.append(item)

            if merged_filters:
                new_compound.append({"filters": merged_filters})

            compound_block["compound"] = new_compound

        return compound_block



EXTRACT_ASSAY_EXPRESSION_FILTERING_CONDITIONS_1_0 = {
    "version": "1.0",
    "output_fields_mapping": {
        "default": [
            {"feature_id": "expression$feature_id"},
            {"sample_id": "expression$sample_id"},
            {"expression": "expression$value"},
        ],
        "additional": [
            {"feature_name": "expr_annotation$gene_name"},
            {"chrom": "expr_annotation$chr"},
            {"start": "expr_annotation$start"},
            {"end": "expr_annotation$end"},
            {"strand": "expr_annotation$strand"},
        ],
    },
    "filtering_conditions": {
        "location": {
            "items_combination_operator": "or",
            "filters_combination_operator": "and",
            "max_item_limit": 10,
            "properties": [
                {
                    "key": "chromosome",
                    "condition": "is",
                    "table_column": "expr_annotation$chr",
                },
                {
                    "keys": ["starting_position", "ending_position"],
                    "condition": "genobin_partial_overlap",
                    "max_range": "250",
                    "table_column": {
                        "starting_position": "expr_annotation$start",
                        "ending_position": "expr_annotation$end",
                    },
                },
            ],
        },
        "annotation": {
            "properties": {
                "feature_id": {
                    "max_item_limit": 100,
                    "condition": "in",
                    "table_column": "expr_annotation$feature_id",
                },
                "feature_name": {
                    "max_item_limit": 100,
                    "condition": "in",
                    "table_column": "expr_annotation$gene_name",
                },
            }
        },
        "expression": {
            "filters_combination_operator": "and",
            "properties": {
                "min_value": {
                    "condition": "greater-than-eq",
                    "table_column": "expression$value",
                },
                "max_value": {
                    "condition": "less-than-eq",
                    "table_column": "expression$value",
                },
            },
        },
        "sample_id": {
            "max_item_limit": 100,
            "condition": "in",
            "table_column": "expression$sample_id",
        },
    },
    "filters_combination_operator": "and",
    "order_by": [
        {"feature_id": "asc"},
        {"sample_id": "asc"},
    ],
}

# EXTRACT_ASSAY_EXPRESSION_FILTERING_CONDITIONS_1_1 does not handle optimization of location filters
EXTRACT_ASSAY_EXPRESSION_FILTERING_CONDITIONS_1_1 = {
    "version": "1.1",
    "output_fields_mapping": {
        "default": [
            {"feature_id": "expression_read_optimized$feature_id"},
            {"sample_id": "expression_read_optimized$sample_id"},
            {"expression": "expression_read_optimized$value"},
        ],
        "additional": [
            {"feature_name": "expression_read_optimized$gene_name"},
            {"chrom": "expression_read_optimized$CHROM"},
            {"start": "expression_read_optimized$start"},
            {"end": "expression_read_optimized$end"},
            {"strand": "expression_read_optimized$strand"},
        ],
    },
    "filtering_conditions": {
        "annotation": {
            "properties": {
                "feature_id": {
                    "max_item_limit": 100,
                    "condition": "in",
                    "table_column": "expression_read_optimized$feature_id",
                },
                "feature_name": {
                    "max_item_limit": 100,
                    "condition": "in",
                    "table_column": "expression_read_optimized$gene_name",
                },
            }
        },
        "expression": {
            "filters_combination_operator": "and",
            "properties": {
                "min_value": {
                    "condition": "greater-than-eq",
                    "table_column": "expression_read_optimized$value",
                },
                "max_value": {
                    "condition": "less-than-eq",
                    "table_column": "expression_read_optimized$value",
                },
            },
        },
        "sample_id": {
            "max_item_limit": 100,
            "condition": "in",
            "table_column": "expression_read_optimized$sample_id",
        },
    },
    "filters_combination_operator": "and",
    "order_by": [
        {"feature_id": "asc"},
        {"sample_id": "asc"},
    ],
}

EXTRACT_ASSAY_EXPRESSION_FILTERING_CONDITIONS_1_1_non_optimized = {
    "version": "1.1",
    "output_fields_mapping": {
        "default": [
            {"feature_id": "expression$feature_id"},
            {"sample_id": "expression$sample_id"},
            {"expression": "expression$value"},
        ],
        "additional": [
            {"feature_name": "expr_annotation$gene_name"},
            {"chrom": "expr_annotation$CHROM"},
            {"start": "expr_annotation$start"},
            {"end": "expr_annotation$end"},
            {"strand": "expr_annotation$strand"},
        ],
    },
    "filtering_conditions": {
        "location": {
            "items_combination_operator": "or",
            "filters_combination_operator": "and",
            "max_item_limit": 10,
            "properties": [
                {
                    "key": "chromosome",
                    "condition": "is",
                    "table_column": "expr_annotation$CHROM",
                },
                {
                    "keys": ["starting_position", "ending_position"],
                    "condition": "genobin_partial_overlap",
                    "max_range": "250",
                    "table_column": {
                        "starting_position": "expr_annotation$start",
                        "ending_position": "expr_annotation$end",
                    },
                },
            ],
        },
        "annotation": {
            "properties": {
                "feature_id": {
                    "max_item_limit": 100,
                    "condition": "in",
                    "table_column": "expr_annotation$feature_id",
                },
                "feature_name": {
                    "max_item_limit": 100,
                    "condition": "in",
                    "table_column": "expr_annotation$gene_name",
                },
            }
        },
        "expression": {
            "filters_combination_operator": "and",
            "properties": {
                "min_value": {
                    "condition": "greater-than-eq",
                    "table_column": "expression$value",
                },
                "max_value": {
                    "condition": "less-than-eq",
                    "table_column": "expression$value",
                },
            },
        },
        "sample_id": {
            "max_item_limit": 100,
            "condition": "in",
            "table_column": "expression$sample_id",
        },
    },
    "filters_combination_operator": "and",
    "order_by": [
        {"feature_id": "asc"},
        {"sample_id": "asc"},
    ],
}



from __future__ import print_function


class ArgsValidator:
    """
    InputsValidator class. Checks for invalid input combinations set by a JSON schema.

    The schema is a dictionary with the following structure:

    {
    "schema_version": "1.0",
    "parser_args":[
                  "path"
                  "...",
                  ]
    "condition_1": {
        "properties": {
            "items": ["path", "json_help"],
        },
        "condition": "at_least_one_required",
        "error_message": {
            "message": "..."
        },
    },
    "condition_2": {
        "properties": {
            "main_key": "path",
            "items": ["output","delim"]
        },
        "condition": "with_at_least_one_required",
        "error_message": {
            "message": "...",
            "type": "warning"
            },

    """

    def __init__(
        self,
        parser_dict,
        schema,
        error_handler=print,
        warning_handler=print,
    ):
        self.parser_dict = parser_dict
        self.schema = schema
        self.error_handler = error_handler
        self.warning_handler = warning_handler

        self.schema_version = schema.get("schema_version")
        self.conditions_funcs = {
            "exclusive": "interpret_exclusive",
            "exclusive_with_exceptions": "interpret_exclusive_with_exceptions",
            "required": "interpret_required",
            "at_least_one_required": "interpret_at_least_one_required",
            "with_at_least_one_required": "interpret_with_at_least_one_required",
            "with_none_of": "interpret_with_none_of",
            "mutually_exclusive_group": "interpret_mutually_exclusive_group",
        }

    ### Schema methods ###
    def validate_schema_conditions(self):
        # Checking if all conditions exist
        present_conditions = [
            value.get("condition")
            for key, value in self.schema.items()
            if key != "schema_version" and key != "parser_args"
        ]
        not_found = set(present_conditions) - set(self.conditions_funcs)
        if len(not_found) != 0:
            self.error_handler("{} schema condition is not defined".format(not_found))

    ### Checking general methods ###
    def interpret_conditions(self):
        for key, value in self.schema.items():
            if key != "schema_version" and key != "parser_args":
                condition = value.get("condition")
                method_to_call = self.conditions_funcs.get(condition)
                getattr(self, method_to_call)(key)

    def throw_schema_message(self, check):
        message_type = self.schema.get(check).get("error_message").get("type")
        message = self.schema.get(check).get("error_message").get("message")
        if message_type == "warning":
            self.throw_warning(message)
        elif isinstance(message_type, type(None)) or (message_type == "error"):
            self.throw_exit_error(message)
        else:
            self.error_handler(
                'Unkown error message in schema: "{}" for key "{}"'.format(type, check)
            )

    def throw_exit_error(self, message):
        self.error_handler(message)

    def throw_warning(self, message):
        self.warning_handler(message)

    def get_parser_values(self, params):
        values = [self.parser_dict.get(p) for p in params]
        return values

    def get_main_key(self, check):
        return self.schema.get(check).get("properties").get("main_key")

    def get_items(self, check):
        return self.schema.get(check).get("properties").get("items")

    def get_items_values(self, check):
        items = self.get_items(check)
        return [self.parser_dict[i] for i in items]

    def get_exceptions(self, check):
        return self.schema.get(check).get("properties").get("exceptions")

    def remove_exceptions_from_list(self, check, list):
        exceptions_list = self.get_exceptions(check)
        for e in exceptions_list:
            list.remove(e)
        return list

    ### Checking specific methods ###
    def interpret_exclusive(self, check):
        self.interpret_exclusive_with_exceptions(check, exception_present=False)

    def interpret_exclusive_with_exceptions(self, check, exception_present=True):
        main_key = self.get_main_key(check)

        # Defining args to check and its values
        args_to_check = self.schema.get("parser_args")[:]
        args_to_check.remove(main_key)
        if exception_present:
            args_to_check = self.remove_exceptions_from_list(check, args_to_check)
        args_to_check_values = self.get_parser_values(args_to_check)

        # True check
        if self.parser_dict.get(main_key) and any(args_to_check_values):
            self.throw_schema_message(check)

    def interpret_with_none_of(self, check):
        main_key = self.get_main_key(check)
        args_to_check_values = self.get_items_values(check)
        if self.parser_dict.get(main_key) and any(args_to_check_values):
            self.throw_schema_message(check)

    def interpret_required(self, check):
        self.interpret_at_least_one_required(check)

    def interpret_at_least_one_required(self, check, main_key=None):
        args_to_check_values = self.get_items_values(check)

        if main_key:
            if self.parser_dict.get(main_key) and not any(args_to_check_values):
                self.throw_schema_message(check)
        else:
            if not any(args_to_check_values):
                self.throw_schema_message(check)

    def interpret_with_at_least_one_required(self, check):
        main_key = self.get_main_key(check)
        self.interpret_at_least_one_required(check, main_key)

    def interpret_mutually_exclusive_group(self, check):
        args_to_check_values = self.get_items_values(check)
        present_args_count = 0

        for arg in args_to_check_values:
            if arg:
                present_args_count += 1

        if present_args_count > 1:
            self.throw_schema_message(check)

    # VALIDATION
    def validate_input_combination(self):
        self.validate_schema_conditions()
        self.interpret_conditions()
